@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2013 Fanout, Inc.
+ * Copyright (C) 2012-2016 Fanout, Inc.
  * 
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,6 +24,24 @@
 class ZhttpResponsePacket
 {
 public:
+	class Id
+	{
+	public:
+		QByteArray id;
+		int seq;
+
+		Id() :
+			seq(-1)
+		{
+		}
+
+		Id(const QByteArray &_id, int _seq = -1) :
+			id(_id),
+			seq(_seq)
+		{
+		}
+	};
+
 	enum Type
 	{
 		Data,
@@ -39,11 +57,10 @@ public:
 	};
 
 	QByteArray from;
-	QByteArray id;
+	QList<Id> ids;
 
 	Type type;
 	QByteArray condition;
-	int seq;
 
 	int credits;
 	bool more;
@@ -57,12 +74,14 @@ public:
 
 	QVariant userData;
 
+	bool multi;
+
 	ZhttpResponsePacket() :
 		type((Type)-1),
-		seq(-1),
 		credits(-1),
 		more(false),
-		code(-1)
+		code(-1),
+		multi(false)
 	{
 	}
 
