@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 Fanout, Inc.
+ * Copyright (C) 2014-2017 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -42,14 +42,17 @@ public:
 			Send,
 			KeepAliveSetup,
 			Close,
-			Detach
+			Detach,
+			Ack
 		};
 
 		QByteArray cid;
 		Type type;
+		QByteArray requestId;
 		QUrl uri;
 		QByteArray contentType;
 		QByteArray message;
+		bool queue;
 		int code;
 		QByteArray route;
 		QByteArray channelPrefix;
@@ -58,6 +61,7 @@ public:
 
 		Item() :
 			type((Type)-1),
+			queue(false),
 			code(-1),
 			ttl(-1),
 			timeout(-1)
