@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014 Fanout, Inc.
+ * Copyright (C) 2014-2016 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -30,6 +30,7 @@
 
 #define IDEAL_CREDITS 200000
 #define SESSION_EXPIRE 60000
+#define KEEPALIVE_INTERVAL 45000
 
 class ZWebSocket::Private : public QObject
 {
@@ -56,6 +57,7 @@ public:
 	QString connectHost;
 	int connectPort;
 	bool ignorePolicies;
+	bool trustConnectHost;
 	bool ignoreTlsErrors;
 	QUrl requestUri;
 	HttpHeaders requestHeaders;
@@ -81,6 +83,7 @@ public:
 	int inSize;
 	int inContentType;
 	int outContentType;
+	bool multi;
 
 	Private(ZWebSocket *_q) :
 		QObject(_q),
@@ -90,6 +93,7 @@ public:
 		state(Idle),
 		connectPort(-1),
 		ignorePolicies(false),
+		trustConnectHost(false),
 		ignoreTlsErrors(false),
 		inSeq(0),
 		outSeq(0),
@@ -105,7 +109,8 @@ public:
 		keepAliveTimer(0),
 		inSize(0),
 		inContentType(-1),
-		outContentType((int)Frame::Text)
+		outContentType((int)Frame::Text),
+		multi(false)
 	{
 		expireTimer = new QTimer(this);
 		connect(expireTimer, &QTimer::timeout, this, &Private::expire_timeout);
@@ -143,12 +148,14 @@ public:
 
 		if(manager)
 		{
+			manager->unregisterKeepAlive(q);
+
 			manager->unlink(q);
 			manager = 0;
 		}
 	}
 
-	bool setupServer(const ZhttpRequestPacket &packet)
+	bool setupServer(int seq, const ZhttpRequestPacket &packet)
 	{
 		if(packet.type != ZhttpRequestPacket::Data)
 		{
@@ -157,7 +164,7 @@ public:
 			return false;
 		}
 
-		if(packet.seq != 0)
+		if(seq != 0)
 		{
 			log_warning("zws server: error, received request with non-zero seq field");
 			writeError("bad-request");
@@ -176,6 +183,9 @@ public:
 		userData = packet.userData;
 		peerAddress = packet.peerAddress;
 
+		if(packet.multi)
+			multi = true;
+
 		return true;
 	}
 
@@ -198,7 +208,28 @@ public:
 
 	void startKeepAlive()
 	{
-		keepAliveTimer->start(SESSION_EXPIRE / 2);
+		if(multi)
+		{
+			if(keepAliveTimer->isActive())
+				keepAliveTimer->stop();
+
+			manager->registerKeepAlive(q);
+		}
+		else
+		{
+			manager->unregisterKeepAlive(q);
+
+			if(!keepAliveTimer->isActive())
+				keepAliveTimer->start(KEEPALIVE_INTERVAL);
+		}
+	}
+
+	void stopKeepAlive()
+	{
+		if(keepAliveTimer->isActive())
+			keepAliveTimer->stop();
+
+		manager->unregisterKeepAlive(q);
 	}
 
 	void refreshTimeout()
@@ -223,6 +254,8 @@ public:
 		out.code = responseCode;
 		out.reason = responseReason;
 		out.headers = responseHeaders;
+		if(multi)
+			out.multi = true;
 		writePacket(out);
 	}
 
@@ -378,13 +411,13 @@ public:
 			inContentType = -1;
 	}
 
-	void handle(const ZhttpRequestPacket &packet)
+	void handle(const QByteArray &id, int seq, const ZhttpRequestPacket &packet)
 	{
 		if(packet.type == ZhttpRequestPacket::Error)
 		{
 			errorCondition = convertError(packet.condition);
 
-			log_debug("zws server: error id=%s cond=%s", packet.id.data(), packet.condition.data());
+			log_debug("zws server: error id=%s cond=%s", id.data(), packet.condition.data());
 
 			state = Idle;
 			cleanup();
@@ -393,7 +426,7 @@ public:
 		}
 		else if(packet.type == ZhttpRequestPacket::Cancel)
 		{
-			log_debug("zws server: received cancel id=%s", packet.id.data());
+			log_debug("zws server: received cancel id=%s", id.data());
 
 			errorCondition = ErrorGeneric;
 			state = Idle;
@@ -402,9 +435,9 @@ public:
 			return;
 		}
 
-		if(packet.seq != inSeq)
+		if(seq != inSeq)
 		{
-			log_warning("zws server: error id=%s received message out of sequence, canceling", packet.id.data());
+			log_warning("zws server: error id=%s received message out of sequence, canceling", id.data());
 
 			tryRespondCancel(packet);
 
@@ -417,12 +450,19 @@ public:
 
 		++inSeq;
 
+		if(!multi && packet.multi)
+		{
+			// switch on multi support
+			multi = true;
+			startKeepAlive(); // re-setup keep alive
+		}
+
 		refreshTimeout();
 
 		if(packet.type == ZhttpRequestPacket::Data || packet.type == ZhttpRequestPacket::Ping || packet.type == ZhttpRequestPacket::Pong)
 		{
 			if(inSize + packet.body.size() > IDEAL_CREDITS)
-				log_warning("zws client: id=%s server is sending too fast", packet.id.data());
+				log_warning("zws client: id=%s server is sending too fast", id.data());
 
 			if(packet.type == ZhttpRequestPacket::Data)
 			{
@@ -472,17 +512,17 @@ public:
 		}
 		else
 		{
-			log_debug("zws server: unsupported packet type id=%s type=%d", packet.id.data(), (int)packet.type);
+			log_debug("zws server: unsupported packet type id=%s type=%d", id.data(), (int)packet.type);
 		}
 	}
 
-	void handle(const ZhttpResponsePacket &packet)
+	void handle(const QByteArray &id, int seq, const ZhttpResponsePacket &packet)
 	{
 		if(packet.type == ZhttpResponsePacket::Error)
 		{
 			errorCondition = convertError(packet.condition);
 
-			log_debug("zws client: error id=%s cond=%s", packet.id.data(), packet.condition.data());
+			log_debug("zws client: error id=%s cond=%s", id.data(), packet.condition.data());
 
 			responseCode = packet.code;
 			responseReason = packet.reason;
@@ -496,7 +536,7 @@ public:
 		}
 		else if(packet.type == ZhttpResponsePacket::Cancel)
 		{
-			log_debug("zws client: received cancel id=%s", packet.id.data());
+			log_debug("zws client: received cancel id=%s", id.data());
 
 			errorCondition = ErrorGeneric;
 			state = Idle;
@@ -508,9 +548,9 @@ public:
 		if(!packet.from.isEmpty())
 			toAddress = packet.from;
 
-		if(packet.seq != inSeq)
+		if(seq != inSeq)
 		{
-			log_warning("zws client: error id=%s received message out of sequence, canceling", packet.id.data());
+			log_warning("zws client: error id=%s received message out of sequence, canceling", id.data());
 
 			tryRespondCancel(packet);
 
@@ -521,11 +561,18 @@ public:
 			return;
 		}
 
-		if(!toAddress.isEmpty() && !keepAliveTimer->isActive())
-			startKeepAlive();
+		if(!toAddress.isEmpty())
+			startKeepAlive(); // only starts if wasn't started already
 
 		++inSeq;
 
+		if(!multi && packet.multi)
+		{
+			// switch on multi support
+			multi = true;
+			startKeepAlive(); // re-setup keep alive
+		}
+
 		refreshTimeout();
 
 		if(state == Connecting)
@@ -535,7 +582,7 @@ public:
 				state = Idle;
 				errorCondition = ErrorGeneric;
 				cleanup();
-				log_warning("zws client: error id=%s initial response wrong type", packet.id.data());
+				log_warning("zws client: error id=%s initial response wrong type", id.data());
 				emit q->error();
 				return;
 			}
@@ -545,7 +592,7 @@ public:
 				state = Idle;
 				errorCondition = ErrorGeneric;
 				cleanup();
-				log_warning("zws client: error id=%s initial ack did not contain from field", packet.id.data());
+				log_warning("zws client: error id=%s initial ack did not contain from field", id.data());
 				emit q->error();
 				return;
 			}
@@ -572,7 +619,7 @@ public:
 			else
 			{
 				if(inSize + packet.body.size() > IDEAL_CREDITS)
-					log_warning("zws client: id=%s server is sending too fast", packet.id.data());
+					log_warning("zws client: id=%s server is sending too fast", id.data());
 
 				if(packet.type == ZhttpResponsePacket::Data)
 				{
@@ -624,7 +671,7 @@ public:
 		}
 		else
 		{
-			log_debug("zws client: unsupported packet type id=%s type=%d", packet.id.data(), (int)packet.type);
+			log_debug("zws client: unsupported packet type id=%s type=%d", id.data(), (int)packet.type);
 		}
 	}
 
@@ -656,12 +703,13 @@ public:
 	{
 		assert(manager);
 
+		bool first = (outSeq == 0);
+
 		ZhttpRequestPacket out = packet;
 		out.from = rid.first;
-		out.id = rid.second;
-		out.seq = outSeq++;
-		
-		if(out.seq == 0)
+		out.ids += ZhttpRequestPacket::Id(rid.second, outSeq++);
+
+		if(first)
 		{
 			manager->writeWs(out);
 		}
@@ -678,8 +726,7 @@ public:
 
 		ZhttpResponsePacket out = packet;
 		out.from = manager->instanceId();
-		out.id = rid.second;
-		out.seq = outSeq++;
+		out.ids += ZhttpResponsePacket::Id(rid.second, outSeq++);
 		out.userData = userData;
 		
 		manager->writeWs(out, rid.first);
@@ -930,9 +977,12 @@ public slots:
 				p.connectPort = connectPort;
 				if(ignorePolicies)
 					p.ignorePolicies = true;
+				if(trustConnectHost)
+					p.trustConnectHost = true;
 				if(ignoreTlsErrors)
 					p.ignoreTlsErrors = true;
 				p.credits = IDEAL_CREDITS;
+				p.multi = true;
 				writePacket(p);
 			}
 			else if(state == Connected || state == ConnectedPeerClosed)
@@ -1005,6 +1055,11 @@ void ZWebSocket::setIgnorePolicies(bool on)
 	d->ignorePolicies = on;
 }
 
+void ZWebSocket::setTrustConnectHost(bool on)
+{
+	d->trustConnectHost = on;
+}
+
 void ZWebSocket::setIgnoreTlsErrors(bool on)
 {
 	d->ignoreTlsErrors = on;
@@ -1156,12 +1211,12 @@ void ZWebSocket::setupClient(ZhttpManager *manager)
 	d->manager->link(this);
 }
 
-bool ZWebSocket::setupServer(ZhttpManager *manager, const ZhttpRequestPacket &packet)
+bool ZWebSocket::setupServer(ZhttpManager *manager, const QByteArray &id, int seq, const ZhttpRequestPacket &packet)
 {
 	d->manager = manager;
 	d->server = true;
-	d->rid = Rid(packet.from, packet.id);
-	return d->setupServer(packet);
+	d->rid = Rid(packet.from, id);
+	return d->setupServer(seq, packet);
 }
 
 void ZWebSocket::startServer()
@@ -1174,18 +1229,28 @@ bool ZWebSocket::isServer() const
 	return d->server;
 }
 
-void ZWebSocket::handle(const ZhttpRequestPacket &packet)
+QByteArray ZWebSocket::toAddress() const
 {
-	assert(d->manager);
-
-	d->handle(packet);
+	return d->toAddress;
 }
 
-void ZWebSocket::handle(const ZhttpResponsePacket &packet)
+int ZWebSocket::outSeqInc()
+{
+	return d->outSeq++;
+}
+
+void ZWebSocket::handle(const QByteArray &id, int seq, const ZhttpRequestPacket &packet)
 {
 	assert(d->manager);
 
-	d->handle(packet);
+	d->handle(id, seq, packet);
+}
+
+void ZWebSocket::handle(const QByteArray &id, int seq, const ZhttpResponsePacket &packet)
+{
+	assert(d->manager);
+
+	d->handle(id, seq, packet);
 }
 
 #include "zwebsocket.moc"
