@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014 Fanout, Inc.
+ * Copyright (C) 2014-2016 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -46,6 +46,7 @@ public:
 	virtual void setConnectHost(const QString &host);
 	virtual void setConnectPort(int port);
 	virtual void setIgnorePolicies(bool on);
+	virtual void setTrustConnectHost(bool on);
 	virtual void setIgnoreTlsErrors(bool on);
 
 	virtual void start(const QUrl &uri, const HttpHeaders &headers);
@@ -78,11 +79,13 @@ private:
 	friend class ZhttpManager;
 	ZWebSocket(QObject *parent = 0);
 	void setupClient(ZhttpManager *manager);
-	bool setupServer(ZhttpManager *manager, const ZhttpRequestPacket &packet);
+	bool setupServer(ZhttpManager *manager, const QByteArray &id, int seq, const ZhttpRequestPacket &packet);
 	void startServer();
 	bool isServer() const;
-	void handle(const ZhttpRequestPacket &packet);
-	void handle(const ZhttpResponsePacket &packet);
+	QByteArray toAddress() const;
+	int outSeqInc();
+	void handle(const QByteArray &id, int seq, const ZhttpRequestPacket &packet);
+	void handle(const QByteArray &id, int seq, const ZhttpResponsePacket &packet);
 };
 
 #endif
