@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2016 Fanout, Inc.
+ * Copyright (C) 2016-2017 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -25,11 +25,14 @@
 class QObject;
 
 class ZrpcManager;
+class StatsPacket;
 class Deferred;
 
 namespace ControlRequest {
 
 Deferred *connCheck(ZrpcManager *controlClient, const CidSet &cids, QObject *parent = 0);
+Deferred *refresh(ZrpcManager *controlClient, const QByteArray &cid, QObject *parent = 0);
+Deferred *report(ZrpcManager *controlClient, const StatsPacket &packet, QObject *parent = 0);
 
 }
 
