@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013-2015 Fanout, Inc.
+ * Copyright (C) 2013-2016 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -24,7 +24,7 @@
 #include <QCommandLineParser>
 #include <QPair>
 #include <QHash>
-#include <QTime>
+#include <QDateTime>
 #include <QTimer>
 #include "qzmqsocket.h"
 #include "qzmqvalve.h"
@@ -39,16 +39,33 @@
 #include "layertracker.h"
 #include "config.h"
 
-#define DEFAULT_HWM 51000
-#define EXPIRE_INTERVAL 1000
+#define DEFAULT_HWM 101000
 #define STATUS_INTERVAL 250
-#define M2_KEEPALIVE_INTERVAL 90000
-#define SESSION_EXPIRE 60000
+#define REFRESH_INTERVAL 1000
+#define M2_CONNECTION_EXPIRE 120000
+#define ZHTTP_EXPIRE 60000
 #define CONTROL_REQUEST_EXPIRE 30000
+#define ZHTTP_CANCEL_RATE 100
+
+#define M2_CONNECTION_SHOULD_PROCESS (M2_CONNECTION_EXPIRE * 3 / 4)
+#define M2_CONNECTION_MUST_PROCESS (M2_CONNECTION_EXPIRE * 4 / 5)
+#define M2_REFRESH_BUCKETS (M2_CONNECTION_SHOULD_PROCESS / REFRESH_INTERVAL)
+
+#define ZHTTP_SHOULD_PROCESS (ZHTTP_EXPIRE * 3 / 4)
+#define ZHTTP_MUST_PROCESS (ZHTTP_EXPIRE * 4 / 5)
+#define ZHTTP_REFRESH_BUCKETS (ZHTTP_SHOULD_PROCESS / REFRESH_INTERVAL)
+
+#define ZHTTP_CANCEL_PER_REFRESH (ZHTTP_CANCEL_RATE * 1000 / REFRESH_INTERVAL)
 
 // make sure this is not larger than Mongrel2's DELIVER_OUTSTANDING_MSGS
 #define M2_PENDING_MAX 16
 
+// make sure this is not larger than Mongrel2's limits.handler_targets
+#define M2_HANDLER_TARGETS_MAX 128
+
+// this doesn't have to match the peer, but we'll set a reasonable number
+#define ZHTTP_IDS_MAX 128
+
 //#define CONTROL_PORT_DEBUG
 
 static void trimlist(QStringList *list)
@@ -281,6 +298,7 @@ public:
 		int packetsPending; // count of packets sent to m2 not yet ack'd
 		Session *session;
 		bool isNew;
+		bool continuation;
 		LayerTracker bodyTracker;
 		LayerTracker packetTracker;
 		QList<M2PendingOutItem> pendingOutItems; // packets yet to send
@@ -289,12 +307,15 @@ public:
 		bool outCreditsEnabled;
 		int outCredits;
 		quint64 subIdBase;
+		qint64 lastRefresh;
+		int refreshBucket;
 
 		M2Connection() :
 			confirmedBytesWritten(0),
 			packetsPending(0),
 			session(0),
 			isNew(false),
+			continuation(false),
 			flowControl(false),
 			waitForAllWritten(false),
 			outCreditsEnabled(false),
@@ -329,13 +350,16 @@ public:
 	{
 	public:
 		Mode mode;
-		int lastActive;
+		qint64 lastActive;
 		QByteArray errorCondition;
 		QByteArray acceptToken; // for websocket
 		bool downClosed; // for websocket
 		bool upClosed; // for websockets
 		QString method;
 		bool responseHeadersOnly; // HEAD, 204, 304
+		qint64 lastRefresh;
+		int refreshBucket;
+		bool pendingCancel;
 
 		// m2 stuff
 		M2Connection *conn;
@@ -357,12 +381,15 @@ public:
 		int inSeq;
 		int pendingInCredits;
 		bool inHandoff;
+		bool multi;
 
 		Session() :
 			lastActive(-1),
 			downClosed(false),
 			upClosed(false),
 			responseHeadersOnly(false),
+			lastRefresh(-1),
+			pendingCancel(false),
 			persistent(false),
 			allowChunked(false),
 			respondKeepAlive(false),
@@ -374,7 +401,8 @@ public:
 			outSeq(0),
 			inSeq(0),
 			pendingInCredits(0),
-			inHandoff(false)
+			inHandoff(false),
+			multi(false)
 		{
 		}
 	};
@@ -399,16 +427,24 @@ public:
 	QHash<Rid, Session*> sessionsByM2Rid;
 	QHash<Rid, Session*> sessionsByZhttpRid;
 	QHash<Rid, Session*> sessionsByZwsRid;
+	QMap<QPair<qint64, M2Connection*>, M2Connection*> m2ConnectionsByLastRefresh;
+	QSet<M2Connection*> m2ConnectionRefreshBuckets[M2_REFRESH_BUCKETS];
+	int currentM2RefreshBucket;
+	QMap<QPair<qint64, Session*>, Session*> sessionsByLastRefresh;
+	QSet<Session*> sessionRefreshBuckets[ZHTTP_REFRESH_BUCKETS];
+	int currentSessionRefreshBucket;
+	QMap<QPair<qint64, Session*>, Session*> sessionsByLastActive;
+	int zhttpCancelMeter;
+	QSet<Session*> sessionsToCancel;
 	int m2_client_buffer;
+	int maxSessions;
 	int zhttpConnectPort;
 	int zwsConnectPort;
 	bool ignorePolicies;
 	QList<ControlPort> controlPorts;
 	QTime time;
-	QTimer *expireTimer;
 	QTimer *statusTimer;
-	QTimer *keepAliveTimer;
-	QTimer *m2KeepAliveTimer;
+	QTimer *refreshTimer;
 
 	Private(App *_q) :
 		QObject(_q),
@@ -423,29 +459,25 @@ public:
 		zws_out_stream_sock(0),
 		m2_in_valve(0),
 		zhttp_in_valve(0),
-		zws_in_valve(0)
+		zws_in_valve(0),
+		currentM2RefreshBucket(0),
+		currentSessionRefreshBucket(0),
+		zhttpCancelMeter(0)
 	{
 		connect(ProcessQuit::instance(), &ProcessQuit::quit, this, &Private::doQuit);
 		connect(ProcessQuit::instance(), &ProcessQuit::hup, this, &Private::reload);
 
-		time.start();
-
-		expireTimer = new QTimer(this);
-		connect(expireTimer, &QTimer::timeout, this, &Private::expire_timeout);
-
 		statusTimer = new QTimer(this);
 		connect(statusTimer, &QTimer::timeout, this, &Private::status_timeout);
 
-		keepAliveTimer = new QTimer(this);
-		connect(keepAliveTimer, &QTimer::timeout, this, &Private::keepAlive_timeout);
-
-		m2KeepAliveTimer = new QTimer(this);
-		connect(m2KeepAliveTimer, &QTimer::timeout, this, &Private::m2KeepAlive_timeout);
+		refreshTimer = new QTimer(this);
+		connect(refreshTimer, &QTimer::timeout, this, &Private::refresh_timeout);
 	}
 
 	~Private()
 	{
-		qDeleteAll(sessionsByM2Rid);
+		qDeleteAll(sessionsByZhttpRid);
+		qDeleteAll(sessionsByZwsRid);
 		qDeleteAll(m2ConnectionsByRid);
 	}
 
@@ -489,17 +521,10 @@ public:
 		if(zws_in_valve)
 			zws_in_valve->open();
 
-		expireTimer->setInterval(EXPIRE_INTERVAL);
-		expireTimer->start();
-
 		statusTimer->setInterval(STATUS_INTERVAL);
-		statusTimer->start();
 
-		keepAliveTimer->setInterval(SESSION_EXPIRE / 2);
-		keepAliveTimer->start();
-
-		m2KeepAliveTimer->setInterval(M2_KEEPALIVE_INTERVAL);
-		m2KeepAliveTimer->start();
+		refreshTimer->setInterval(REFRESH_INTERVAL);
+		refreshTimer->start();
 
 		log_info("started");
 	}
@@ -570,6 +595,8 @@ public:
 		if(m2_client_buffer <= 0)
 			m2_client_buffer = 200000;
 
+		maxSessions = settings.value("max_open_requests", -1).toInt();
+
 		m2_send_idents.clear();
 		foreach(const QString &s, str_m2_send_idents)
 			m2_send_idents += s.toUtf8();
@@ -620,6 +647,7 @@ public:
 		connect(m2_in_valve, &QZmq::Valve::readyRead, this, &Private::m2_in_readyRead);
 
 		m2_out_sock = new QZmq::Socket(QZmq::Socket::Pub, this);
+		m2_out_sock->setShutdownWaitTime(0);
 		m2_out_sock->setHwm(DEFAULT_HWM);
 		m2_out_sock->setWriteQueueEnabled(false);
 		foreach(const QString &spec, m2_out_specs)
@@ -764,6 +792,7 @@ public:
 			}
 
 			zws_out_stream_sock = new QZmq::Socket(QZmq::Socket::Router, this);
+			zws_out_stream_sock->setShutdownWaitTime(0);
 			zws_out_stream_sock->setHwm(DEFAULT_HWM);
 			if(zws_connect)
 			{
@@ -787,6 +816,13 @@ public:
 		return true;
 	}
 
+	void removeConnection(M2Connection *conn)
+	{
+		m2ConnectionRefreshBuckets[conn->refreshBucket].remove(conn);
+		m2ConnectionsByLastRefresh.remove(QPair<qint64, M2Connection*>(conn->lastRefresh, conn));
+		m2ConnectionsByRid.remove(Rid(m2_send_idents[conn->identIndex], conn->id));
+	}
+
 	void unlinkConnection(Session *s)
 	{
 		if(s->conn)
@@ -799,16 +835,78 @@ public:
 		}
 	}
 
-	void destroySession(Session *s)
+	int smallestM2RefreshBucket()
+	{
+		int best = -1;
+		int bestSize;
+		for(int n = 0; n < M2_REFRESH_BUCKETS; ++n)
+		{
+			if(best == -1 || m2ConnectionRefreshBuckets[n].count() < bestSize)
+			{
+				best = n;
+				bestSize = m2ConnectionRefreshBuckets[n].count();
+			}
+		}
+
+		return best;
+	}
+
+	int smallestSessionRefreshBucket()
+	{
+		int best = -1;
+		int bestSize;
+		for(int n = 0; n < ZHTTP_REFRESH_BUCKETS; ++n)
+		{
+			if(best == -1 || sessionRefreshBuckets[n].count() < bestSize)
+			{
+				best = n;
+				bestSize = sessionRefreshBuckets[n].count();
+			}
+		}
+
+		return best;
+	}
+
+	void removeSession(Session *s)
 	{
 		unlinkConnection(s);
+
+		sessionsToCancel -= s;
+
+		if(s->lastRefresh >= 0)
+		{
+			QPair<qint64, Session*> k(s->lastRefresh, s);
+			if(sessionsByLastRefresh.contains(k))
+			{
+				sessionRefreshBuckets[s->refreshBucket].remove(s);
+				sessionsByLastRefresh.remove(k);
+			}
+		}
+
+		sessionsByLastActive.remove(QPair<qint64, Session*>(s->lastActive, s));
+
 		if(s->mode == Http)
 			sessionsByZhttpRid.remove(Rid(zhttpInstanceId, s->id));
 		else // WebSocket
 			sessionsByZwsRid.remove(Rid(zwsInstanceId, s->id));
+	}
+
+	void destroySession(Session *s)
+	{
+		removeSession(s);
 		delete s;
 	}
 
+	void queueCancelSession(Session *s)
+	{
+		assert(!s->zhttpAddress.isEmpty());
+
+		unlinkConnection(s);
+
+		s->pendingCancel = true;
+		sessionsToCancel += s;
+	}
+
 	void destroySessionAndErrorConnection(Session *s)
 	{
 		M2Connection *conn = s->conn;
@@ -821,7 +919,8 @@ public:
 	{
 		QByteArray buf = packet.toByteArray();
 
-		log_debug("m2: OUT [%s]", buf.data());
+		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
+			log_debug("m2: OUT [%s]", buf.mid(0, 1000).data());
 
 		m2_out_sock->write(QList<QByteArray>() << buf);
 	}
@@ -856,6 +955,22 @@ public:
 		m2_out_write(mresp);
 	}
 
+	void m2_writeCtlMany(const QByteArray &sender, const QList<QByteArray> &connIds, const QVariant &args)
+	{
+		assert(!connIds.isEmpty());
+
+		M2ResponsePacket mresp;
+		mresp.sender = sender;
+		mresp.id = "X";
+		foreach(const QByteArray &id, connIds)
+			mresp.id += " " + id;
+		QVariantList parts;
+		parts += QByteArray("ctl");
+		parts += args;
+		mresp.data = TnetString::fromVariant(parts);
+		m2_out_write(mresp);
+	}
+
 	void m2_writeCtlCancel(const QByteArray &sender, const QByteArray &id)
 	{
 		M2ResponsePacket mresp;
@@ -873,7 +988,7 @@ public:
 	void m2_writeCtlCancel(M2Connection *conn)
 	{
 		m2_writeCtlCancel(m2_send_idents[conn->identIndex], conn->id);
-		m2ConnectionsByRid.remove(Rid(m2_send_idents[conn->identIndex], conn->id));
+		removeConnection(conn);
 		delete conn;
 	}
 
@@ -1065,7 +1180,7 @@ public:
 	void m2_writeClose(M2Connection *conn)
 	{
 		m2_writeClose(m2_send_idents[conn->identIndex], conn->id);
-		m2ConnectionsByRid.remove(Rid(m2_send_idents[conn->identIndex], conn->id));
+		removeConnection(conn);
 		delete conn;
 	}
 
@@ -1087,7 +1202,8 @@ public:
 
 		QByteArray buf = QByteArray("T") + TnetString::fromVariant(packet.toVariant());
 
-		log_debug("%s: OUT %s", logprefix, buf.mid(0, 1000).data());
+		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
+			log_debug("%s: OUT %s", logprefix, buf.mid(0, 1000).data());
 
 		if(mode == Http)
 			zhttp_out_sock->write(QList<QByteArray>() << buf);
@@ -1101,7 +1217,8 @@ public:
 
 		QByteArray buf = QByteArray("T") + TnetString::fromVariant(packet.toVariant());
 
-		log_debug("%s: OUT instance=%s %s", logprefix, instanceAddress.data(), buf.mid(0, 1000).data());
+		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
+			log_debug("%s: OUT instance=%s %s", logprefix, instanceAddress.data(), buf.mid(0, 1000).data());
 
 		QList<QByteArray> message;
 		message += instanceAddress;
@@ -1118,8 +1235,7 @@ public:
 	{
 		ZhttpRequestPacket out = packet;
 		out.from = (s->mode == Http ? zhttpInstanceId : zwsInstanceId);
-		out.id = s->id;
-		out.seq = (s->outSeq)++;
+		out.ids += ZhttpRequestPacket::Id(s->id, (s->outSeq)++);
 		zhttp_out_write(s->mode, out);
 	}
 
@@ -1129,15 +1245,15 @@ public:
 
 		ZhttpRequestPacket out = packet;
 		out.from = (s->mode == Http ? zhttpInstanceId : zwsInstanceId);
-		out.id = s->id;
-		out.seq = (s->outSeq)++;
+		out.ids += ZhttpRequestPacket::Id(s->id, (s->outSeq)++);
 		zhttp_out_write(s->mode, out, s->zhttpAddress);
 	}
 
 	void handleControlResponse(int index, const QVariant &data)
 	{
 #ifdef CONTROL_PORT_DEBUG
-		log_debug("m2: IN control %s %s", m2_send_idents[index].data(), qPrintable(TnetString::variantToString(data)));
+		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
+			log_debug("m2: IN control %s %s", m2_send_idents[index].data(), qPrintable(TnetString::variantToString(data)));
 #endif
 
 		if(data.type() != QVariant::Hash)
@@ -1215,7 +1331,7 @@ public:
 			if(conn->session)
 				endSession(conn->session, "disconnected");
 
-			m2ConnectionsByRid.remove(Rid(m2_send_idents[conn->identIndex], conn->id));
+			removeConnection(conn);
 			delete conn;
 		}
 	}
@@ -1233,8 +1349,15 @@ public:
 
 		if(conn->session && bodyWritten > 0)
 		{
-			conn->session->lastActive = time.elapsed();
-			handleSessionBodyWritten(conn->session, bodyWritten, giveCredits);
+			Session *s = conn->session;
+
+			// update lastActive
+			qint64 now = QDateTime::currentMSecsSinceEpoch();
+			sessionsByLastActive.remove(QPair<qint64, Session*>(s->lastActive, s));
+			s->lastActive = now;
+			sessionsByLastActive.insert(QPair<qint64, Session*>(s->lastActive, s), s);
+
+			handleSessionBodyWritten(s, bodyWritten, giveCredits);
 		}
 	}
 
@@ -1271,19 +1394,28 @@ public:
 		}
 		else
 		{
-			ZhttpRequestPacket zreq;
-
-			if(!errorCondition.isEmpty())
+			if(sessionsToCancel.isEmpty() && zhttpCancelMeter < ZHTTP_CANCEL_PER_REFRESH)
 			{
-				zreq.type = ZhttpRequestPacket::Error;
-				zreq.condition = "disconnected";
+				++zhttpCancelMeter;
+
+				ZhttpRequestPacket zreq;
+
+				if(!errorCondition.isEmpty())
+				{
+					zreq.type = ZhttpRequestPacket::Error;
+					zreq.condition = "disconnected";
+				}
+				else
+					zreq.type = ZhttpRequestPacket::Cancel;
+
+				zhttp_out_write(s, zreq);
+
+				destroySession(s);
 			}
 			else
-				zreq.type = ZhttpRequestPacket::Cancel;
-
-			zhttp_out_write(s, zreq);
-
-			destroySession(s);
+			{
+				queueCancelSession(s);
+			}
 		}
 	}
 
@@ -1318,7 +1450,8 @@ public:
 			return;
 		}
 
-		log_debug("%s: IN %s", logprefix, dataRaw.data());
+		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
+			log_debug("%s: IN %s", logprefix, dataRaw.mid(0, 1000).data());
 
 		ZhttpResponsePacket zresp;
 		if(!zresp.fromVariant(data))
@@ -1327,11 +1460,17 @@ public:
 			return;
 		}
 
+		foreach(const ZhttpResponsePacket::Id &id, zresp.ids)
+			handleZhttpIn(logprefix, mode, id.id, id.seq, zresp);
+	}
+
+	void handleZhttpIn(const char *logprefix, Mode mode, const QByteArray &id, int seq, const ZhttpResponsePacket &zresp)
+	{
 		Session *s;
 		if(mode == Http)
-			s = sessionsByZhttpRid.value(Rid(zhttpInstanceId, zresp.id));
+			s = sessionsByZhttpRid.value(Rid(zhttpInstanceId, id));
 		else // WebSocket
-			s = sessionsByZwsRid.value(Rid(zwsInstanceId, zresp.id));
+			s = sessionsByZwsRid.value(Rid(zwsInstanceId, id));
 
 		if(!s)
 		{
@@ -1342,7 +1481,7 @@ public:
 			{
 				ZhttpRequestPacket zreq;
 				zreq.from = (mode == Http ? zhttpInstanceId : zwsInstanceId);
-				zreq.id = zresp.id;
+				zreq.ids += ZhttpRequestPacket::Id(id);
 				zreq.type = ZhttpRequestPacket::Cancel;
 				zhttp_out_write(mode, zreq, zresp.from);
 			}
@@ -1368,7 +1507,7 @@ public:
 
 				s->zhttpAddress = zresp.from;
 
-				if(zresp.seq != 0)
+				if(seq != 0)
 				{
 					log_warning("%s: received first response of sequence without valid seq, canceling", logprefix);
 					ZhttpRequestPacket zreq;
@@ -1385,9 +1524,9 @@ public:
 					s->zhttpAddress = zresp.from;
 
 				// if not sequenced, but seq is provided, then it must be 0
-				if(zresp.seq != -1 && zresp.seq != 0)
+				if(seq != -1 && seq != 0)
 				{
-					log_warning("%s: received response out of sequence (got=%d, expected=-1,0), canceling", logprefix, zresp.seq);
+					log_warning("%s: received response out of sequence (got=%d, expected=-1,0), canceling", logprefix, seq);
 
 					if(!s->zhttpAddress.isEmpty())
 					{
@@ -1403,9 +1542,9 @@ public:
 		}
 		else
 		{
-			if(zresp.seq != -1 && zresp.seq != s->inSeq)
+			if(seq != -1 && seq != s->inSeq)
 			{
-				log_warning("%s: received response out of sequence (got=%d, expected=%d), canceling", logprefix, zresp.seq, s->inSeq);
+				log_warning("%s: received response out of sequence (got=%d, expected=%d), canceling", logprefix, seq, s->inSeq);
 				ZhttpRequestPacket zreq;
 				zreq.type = ZhttpRequestPacket::Cancel;
 				zhttp_out_write(s, zreq);
@@ -1419,10 +1558,29 @@ public:
 		}
 
 		// only bump sequence if seq was provided
-		if(zresp.seq != -1)
+		if(seq != -1)
 			++(s->inSeq);
 
-		s->lastActive = time.elapsed();
+		qint64 now = QDateTime::currentMSecsSinceEpoch();
+
+		if(s->lastRefresh < 0 && !s->zhttpAddress.isEmpty())
+		{
+			// once we have the peer's address, set up refresh
+
+			s->lastRefresh = now;
+			sessionsByLastRefresh.insert(QPair<qint64, Session*>(s->lastRefresh, s), s);
+
+			s->refreshBucket = smallestSessionRefreshBucket();
+			sessionRefreshBuckets[s->refreshBucket] += s;
+		}
+
+		// update lastActive
+		sessionsByLastActive.remove(QPair<qint64, Session*>(s->lastActive, s));
+		s->lastActive = now;
+		sessionsByLastActive.insert(QPair<qint64, Session*>(s->lastActive, s), s);
+
+		if(s->pendingCancel)
+			return;
 
 		// a session without a connection is just waiting to report error
 		if(!s->conn)
@@ -1446,15 +1604,39 @@ public:
 			return;
 		}
 
+		// if peer supports multi feature then flag it on the session
+		bool multiWasTurnedOn = false;
+		if(!s->multi && zresp.multi)
+		{
+			s->multi = true;
+			multiWasTurnedOn = true;
+		}
+
 		if(s->inHandoff)
 		{
 			// receiving any message means handoff is complete
 			s->inHandoff = false;
 
+			// refresh would have already been set up once if we are here
+			assert(s->lastRefresh >= 0);
+
+			sessionsByLastRefresh.insert(QPair<qint64, Session*>(s->lastRefresh, s), s);
+			s->refreshBucket = smallestSessionRefreshBucket();
+			sessionRefreshBuckets[s->refreshBucket] += s;
+
 			// in order to have been in a handoff state, we would have
 			//   had to receive a from address sometime earlier, so it
 			//   should be safe to call zhttp_out_write with session.
 
+			if(multiWasTurnedOn)
+			{
+				// acknowledge the feature
+				ZhttpRequestPacket zreq;
+				zreq.type = ZhttpRequestPacket::KeepAlive;
+				zreq.multi = true;
+				zhttp_out_write(s, zreq);
+			}
+
 			if(s->mode == Http)
 			{
 				if(!s->pendingIn.isEmpty())
@@ -1669,10 +1851,19 @@ public:
 				else
 				{
 					int opcode;
-					if(zresp.contentType == "binary")
-						opcode = 2;
-					else // text
-						opcode = 1;
+					if(s->conn->continuation)
+					{
+						opcode = 0;
+					}
+					else
+					{
+						if(zresp.contentType == "binary")
+							opcode = 2;
+						else // text
+							opcode = 1;
+					}
+
+					s->conn->continuation = zresp.more;
 
 					QByteArray frame = makeWsHeader(!zresp.more, opcode, zresp.body.size()) + zresp.body;
 
@@ -1738,6 +1929,12 @@ public:
 		{
 			s->inHandoff = true;
 
+			sessionRefreshBuckets[s->refreshBucket].remove(s);
+			sessionsByLastRefresh.remove(QPair<qint64, Session*>(s->lastRefresh, s));
+
+			// whoever picks up after handoff can turn this on
+			s->multi = false;
+
 			ZhttpRequestPacket zreq;
 			zreq.type = ZhttpRequestPacket::HandoffProceed;
 			zhttp_out_write(s, zreq);
@@ -1779,6 +1976,324 @@ public:
 		}
 	}
 
+	void refreshM2Connections(qint64 now)
+	{
+		QHash<int, QList<QByteArray> > connIdListBySender;
+
+		// process the current bucket
+		const QSet<M2Connection*> &bucket = m2ConnectionRefreshBuckets[currentM2RefreshBucket];
+		foreach(M2Connection *conn, bucket)
+		{
+			// move to the end
+			QPair<qint64, M2Connection*> k(conn->lastRefresh, conn);
+			m2ConnectionsByLastRefresh.remove(k);
+			conn->lastRefresh = now;
+			m2ConnectionsByLastRefresh.insert(QPair<qint64, M2Connection*>(conn->lastRefresh, conn), conn);
+
+			if(!connIdListBySender.contains(conn->identIndex))
+				connIdListBySender.insert(conn->identIndex, QList<QByteArray>());
+
+			QList<QByteArray> &connIdList = connIdListBySender[conn->identIndex];
+			connIdList += conn->id;
+
+			// if we're at max, send out now
+			if(connIdList.count() >= M2_HANDLER_TARGETS_MAX)
+			{
+				QVariantHash args;
+				args["keep-alive"] = true;
+				m2_writeCtlMany(m2_send_idents[conn->identIndex], connIdList, args);
+
+				connIdList.clear();
+				connIdListBySender.remove(conn->identIndex);
+			}
+		}
+
+		// process any others
+		qint64 threshold = now - M2_CONNECTION_MUST_PROCESS;
+		while(!m2ConnectionsByLastRefresh.isEmpty())
+		{
+			QMap<QPair<qint64, M2Connection*>, M2Connection*>::iterator it = m2ConnectionsByLastRefresh.begin();
+			M2Connection *conn = it.value();
+
+			if(conn->lastRefresh > threshold)
+				break;
+
+			// move to the end
+			m2ConnectionsByLastRefresh.erase(it);
+			conn->lastRefresh = now;
+			m2ConnectionsByLastRefresh.insert(QPair<qint64, M2Connection*>(conn->lastRefresh, conn), conn);
+
+			if(!connIdListBySender.contains(conn->identIndex))
+				connIdListBySender.insert(conn->identIndex, QList<QByteArray>());
+
+			QList<QByteArray> &connIdList = connIdListBySender[conn->identIndex];
+			connIdList += conn->id;
+
+			// if we're at max, send out now
+			if(connIdList.count() >= M2_HANDLER_TARGETS_MAX)
+			{
+				QVariantHash args;
+				args["keep-alive"] = true;
+				m2_writeCtlMany(m2_send_idents[conn->identIndex], connIdList, args);
+
+				connIdList.clear();
+				connIdListBySender.remove(conn->identIndex);
+			}
+		}
+
+		// send last packet
+		QHashIterator<int, QList<QByteArray> > cit(connIdListBySender);
+		while(cit.hasNext())
+		{
+			cit.next();
+			int index = cit.key();
+			const QList<QByteArray> &connIdList = cit.value();
+
+			if(!connIdList.isEmpty())
+			{
+				QVariantHash args;
+				args["keep-alive"] = true;
+				m2_writeCtlMany(m2_send_idents[index], connIdList, args);
+			}
+		}
+
+		++currentM2RefreshBucket;
+		if(currentM2RefreshBucket >= M2_REFRESH_BUCKETS)
+			currentM2RefreshBucket = 0;
+	}
+
+	void refreshSessions(qint64 now)
+	{
+		QHash<QByteArray, QList<Session*> > sessionListBySender[2]; // index corresponds to mode
+
+		// process the current bucket
+		const QSet<Session*> &bucket = sessionRefreshBuckets[currentSessionRefreshBucket];
+		foreach(Session *s, bucket)
+		{
+			assert(!s->inHandoff && !s->zhttpAddress.isEmpty());
+
+			// move to the end
+			QPair<qint64, Session*> k(s->lastRefresh, s);
+			sessionsByLastRefresh.remove(k);
+			s->lastRefresh = now;
+			sessionsByLastRefresh.insert(QPair<qint64, Session*>(s->lastRefresh, s), s);
+
+			if(s->multi)
+			{
+				if(!sessionListBySender[s->mode].contains(s->zhttpAddress))
+					sessionListBySender[s->mode].insert(s->zhttpAddress, QList<Session*>());
+
+				QList<Session*> &sessionList = sessionListBySender[s->mode][s->zhttpAddress];
+				sessionList += s;
+
+				// if we're at max, send out now
+				if(sessionList.count() >= ZHTTP_IDS_MAX)
+				{
+					ZhttpRequestPacket zreq;
+					zreq.from = (s->mode == Http ? zhttpInstanceId : zwsInstanceId);
+					foreach(Session *i, sessionList)
+						zreq.ids += ZhttpRequestPacket::Id(i->id, (i->outSeq)++);
+					zreq.type = ZhttpRequestPacket::KeepAlive;
+					zhttp_out_write(s->mode, zreq, s->zhttpAddress);
+
+					sessionList.clear();
+					sessionListBySender[s->mode].remove(s->zhttpAddress);
+				}
+			}
+			else
+			{
+				// session doesn't support sending with multiple ids
+				ZhttpRequestPacket zreq;
+				zreq.from = (s->mode == Http ? zhttpInstanceId : zwsInstanceId);
+				zreq.ids += ZhttpRequestPacket::Id(s->id, (s->outSeq)++);
+				zreq.type = ZhttpRequestPacket::KeepAlive;
+				zhttp_out_write(s->mode, zreq, s->zhttpAddress);
+			}
+		}
+
+		// process any others
+		qint64 threshold = now - ZHTTP_MUST_PROCESS;
+		while(!sessionsByLastRefresh.isEmpty())
+		{
+			QMap<QPair<qint64, Session*>, Session*>::iterator it = sessionsByLastRefresh.begin();
+			Session *s = it.value();
+
+			if(s->lastRefresh > threshold)
+				break;
+
+			assert(!s->inHandoff && !s->zhttpAddress.isEmpty());
+
+			// move to the end
+			sessionsByLastRefresh.erase(it);
+			s->lastRefresh = now;
+			sessionsByLastRefresh.insert(QPair<qint64, Session*>(s->lastRefresh, s), s);
+
+			if(s->multi)
+			{
+				if(!sessionListBySender[s->mode].contains(s->zhttpAddress))
+					sessionListBySender[s->mode].insert(s->zhttpAddress, QList<Session*>());
+
+				QList<Session*> &sessionList = sessionListBySender[s->mode][s->zhttpAddress];
+				sessionList += s;
+
+				// if we're at max, send out now
+				if(sessionList.count() >= ZHTTP_IDS_MAX)
+				{
+					ZhttpRequestPacket zreq;
+					zreq.from = (s->mode == Http ? zhttpInstanceId : zwsInstanceId);
+					foreach(Session *i, sessionList)
+						zreq.ids += ZhttpRequestPacket::Id(i->id, (i->outSeq)++);
+					zreq.type = ZhttpRequestPacket::KeepAlive;
+					zhttp_out_write(s->mode, zreq, s->zhttpAddress);
+
+					sessionList.clear();
+					sessionListBySender[s->mode].remove(s->zhttpAddress);
+				}
+			}
+			else
+			{
+				// session doesn't support sending with multiple ids
+				ZhttpRequestPacket zreq;
+				zreq.from = (s->mode == Http ? zhttpInstanceId : zwsInstanceId);
+				zreq.ids += ZhttpRequestPacket::Id(s->id, (s->outSeq)++);
+				zreq.type = ZhttpRequestPacket::KeepAlive;
+				zhttp_out_write(s->mode, zreq, s->zhttpAddress);
+			}
+		}
+
+		// send last packets
+		for(int n = 0; n < 2; ++n)
+		{
+			Mode mode = (Mode)n;
+
+			QHashIterator<QByteArray, QList<Session*> > sit(sessionListBySender[n]);
+			while(sit.hasNext())
+			{
+				sit.next();
+				const QByteArray &zhttpAddress = sit.key();
+				const QList<Session*> &sessionList = sit.value();
+
+				if(!sessionList.isEmpty())
+				{
+					ZhttpRequestPacket zreq;
+					zreq.from = (mode == Http ? zhttpInstanceId : zwsInstanceId);
+					foreach(Session *s, sessionList)
+						zreq.ids += ZhttpRequestPacket::Id(s->id, (s->outSeq)++);
+					zreq.type = ZhttpRequestPacket::KeepAlive;
+					zhttp_out_write(mode, zreq, zhttpAddress);
+				}
+			}
+		}
+
+		++currentSessionRefreshBucket;
+		if(currentSessionRefreshBucket >= ZHTTP_REFRESH_BUCKETS)
+			currentSessionRefreshBucket = 0;
+	}
+
+	void expireSessions(qint64 now)
+	{
+		qint64 threshold = now - ZHTTP_EXPIRE;
+		while(!sessionsByLastActive.isEmpty())
+		{
+			QMap<QPair<qint64, Session*>, Session*>::iterator it = sessionsByLastActive.begin();
+			Session *s = it.value();
+
+			if(s->lastActive > threshold)
+				break;
+
+			log_warning("timing out request %s", s->id.data());
+			destroySessionAndErrorConnection(s);
+		}
+	}
+
+	void cancelSessions()
+	{
+		int sent = zhttpCancelMeter;
+
+		if(zhttpCancelMeter > ZHTTP_CANCEL_PER_REFRESH)
+			zhttpCancelMeter -= ZHTTP_CANCEL_PER_REFRESH;
+		else
+			zhttpCancelMeter = 0;
+
+		QHash<QByteArray, QList<Session*> > sessionListBySender[2]; // index corresponds to mode
+
+		while(!sessionsToCancel.isEmpty() && sent < ZHTTP_CANCEL_PER_REFRESH)
+		{
+			QSet<Session*>::iterator it = sessionsToCancel.begin();
+			Session *s = (*it);
+			sessionsToCancel.erase(it);
+
+			if(s->multi)
+			{
+				if(!sessionListBySender[s->mode].contains(s->zhttpAddress))
+					sessionListBySender[s->mode].insert(s->zhttpAddress, QList<Session*>());
+
+				QList<Session*> &sessionList = sessionListBySender[s->mode][s->zhttpAddress];
+				sessionList += s;
+
+				// if we're at max, send out now
+				if(sessionList.count() >= ZHTTP_IDS_MAX)
+				{
+					ZhttpRequestPacket zreq;
+					zreq.from = (s->mode == Http ? zhttpInstanceId : zwsInstanceId);
+					foreach(Session *i, sessionList)
+						zreq.ids += ZhttpRequestPacket::Id(i->id, (i->outSeq)++);
+					zreq.type = ZhttpRequestPacket::Cancel;
+					zhttp_out_write(s->mode, zreq, s->zhttpAddress);
+
+					Mode mode = s->mode;
+					QByteArray zhttpAddress = s->zhttpAddress;
+
+					foreach(Session *s, sessionList)
+						destroySession(s);
+
+					sessionList.clear();
+					sessionListBySender[mode].remove(zhttpAddress);
+				}
+			}
+			else
+			{
+				// session doesn't support sending with multiple ids
+				ZhttpRequestPacket zreq;
+				zreq.from = (s->mode == Http ? zhttpInstanceId : zwsInstanceId);
+				zreq.ids += ZhttpRequestPacket::Id(s->id, (s->outSeq)++);
+				zreq.type = ZhttpRequestPacket::Cancel;
+				zhttp_out_write(s->mode, zreq, s->zhttpAddress);
+
+				destroySession(s);
+			}
+
+			++sent;
+		}
+
+		// send last packets
+		for(int n = 0; n < 2; ++n)
+		{
+			Mode mode = (Mode)n;
+
+			QHashIterator<QByteArray, QList<Session*> > sit(sessionListBySender[n]);
+			while(sit.hasNext())
+			{
+				sit.next();
+				const QByteArray &zhttpAddress = sit.key();
+				const QList<Session*> &sessionList = sit.value();
+
+				if(!sessionList.isEmpty())
+				{
+					ZhttpRequestPacket zreq;
+					zreq.from = (mode == Http ? zhttpInstanceId : zwsInstanceId);
+					foreach(Session *s, sessionList)
+						zreq.ids += ZhttpRequestPacket::Id(s->id, (s->outSeq)++);
+					zreq.type = ZhttpRequestPacket::Cancel;
+					zhttp_out_write(mode, zreq, zhttpAddress);
+				}
+
+				foreach(Session *s, sessionList)
+					destroySession(s);
+			}
+		}
+	}
+
 private slots:
 	void m2_in_readyRead(const QList<QByteArray> &message)
 	{
@@ -1788,7 +2303,8 @@ private slots:
 			return;
 		}
 
-		log_debug("m2: IN %s", message[0].mid(0, 1000).data());
+		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
+			log_debug("m2: IN %s", message[0].mid(0, 1000).data());
 
 		M2RequestPacket mreq;
 		if(!mreq.fromByteArray(message[0]))
@@ -1810,12 +2326,14 @@ private slots:
 			if(conn->session)
 				endSession(conn->session);
 
-			m2ConnectionsByRid.remove(rid);
+			removeConnection(conn);
 			delete conn;
 
 			return;
 		}
 
+		qint64 now = QDateTime::currentMSecsSinceEpoch();
+
 		Rid m2Rid(mreq.sender, mreq.id);
 
 		M2Connection *conn = m2ConnectionsByRid.value(m2Rid);
@@ -1876,6 +2394,9 @@ private slots:
 				{
 					log_debug("activating control port index=%d", index);
 					controlPorts[index].state = ControlPort::Idle;
+
+					if(!statusTimer->isActive())
+						statusTimer->start();
 				}
 
 				// if we were in the middle of requesting control info when this
@@ -1891,6 +2412,12 @@ private slots:
 			}
 
 			m2ConnectionsByRid.insert(m2Rid, conn);
+
+			conn->lastRefresh = now;
+			m2ConnectionsByLastRefresh.insert(QPair<qint64, M2Connection*>(conn->lastRefresh, conn), conn);
+
+			conn->refreshBucket = smallestM2RefreshBucket();
+			m2ConnectionRefreshBuckets[conn->refreshBucket] += conn;
 		}
 		else
 		{
@@ -1952,10 +2479,6 @@ private slots:
 			if(host.isEmpty())
 				host = "localhost";
 
-			int at = host.indexOf(':');
-			if(at != -1)
-				host = host.mid(0, at);
-
 			if(!validateHost(host))
 			{
 				log_warning("m2: invalid host [%s]", host.data());
@@ -1979,10 +2502,16 @@ private slots:
 				return;
 			}
 
+			if(maxSessions >= 0 && sessionsByZhttpRid.count() + sessionsByZwsRid.count() >= maxSessions)
+			{
+				log_warning("m2: max open sessions reached (%d), refusing new session", maxSessions);
+				m2_writeErrorClose(conn);
+				return;
+			}
+
 			s = new Session;
 			s->conn = conn;
 			s->conn->session = s;
-			s->lastActive = time.elapsed();
 			s->id = m2_send_idents[conn->identIndex] + '_' + conn->id + '_' + QByteArray::number((conn->subIdBase)++, 16);
 			s->method = mreq.method;
 
@@ -2021,6 +2550,11 @@ private slots:
 
 			sessionsByM2Rid.insert(m2Rid, s);
 
+			qint64 now = QDateTime::currentMSecsSinceEpoch();
+
+			s->lastActive = now;
+			sessionsByLastActive.insert(QPair<qint64, Session*>(s->lastActive, s), s);
+
 			if(mreq.type == M2RequestPacket::HttpRequest)
 				sessionsByZhttpRid.insert(Rid(zhttpInstanceId, s->id), s);
 			else // WebSocketHandshake
@@ -2053,6 +2587,8 @@ private slots:
 				zreq.more = !s->inFinished;
 			}
 
+			zreq.multi = true;
+
 			zhttp_out_writeFirst(s, zreq);
 		}
 		else
@@ -2228,6 +2764,18 @@ private slots:
 			{
 				log_debug("deactivating control port index=%d", index);
 				c.state = ControlPort::Disabled;
+
+				bool allDisabled = true;
+				foreach(const ControlPort &i, controlPorts)
+				{
+					if(i.state != ControlPort::Disabled)
+					{
+						allDisabled = false;
+						break;
+					}
+				}
+				if(allDisabled)
+					statusTimer->stop();
 			}
 
 			c.reqStartTime = -1;
@@ -2244,25 +2792,6 @@ private slots:
 		handleZhttpIn(WebSocket, message);
 	}
 
-	void expire_timeout()
-	{
-		int now = time.elapsed();
-		QList<Session*> toDelete;
-		QHashIterator<Rid, Session*> it(sessionsByM2Rid);
-		while(it.hasNext())
-		{
-			it.next();
-			Session *s = it.value();
-			if(s->lastActive + SESSION_EXPIRE <= now)
-				toDelete += s;
-		}
-		foreach(Session *s, toDelete)
-		{
-			log_warning("timing out request %s", s->id.data());
-			destroySessionAndErrorConnection(s);
-		}
-	}
-
 	void status_timeout()
 	{
 		int now = time.elapsed();
@@ -2287,53 +2816,14 @@ private slots:
 		}
 	}
 
-	void keepAlive_timeout()
+	void refresh_timeout()
 	{
-		{
-			QHashIterator<Rid, Session*> it(sessionsByZhttpRid);
-			while(it.hasNext())
-			{
-				it.next();
-				Session *s = it.value();
+		qint64 now = QDateTime::currentMSecsSinceEpoch();
 
-				if(!s->inHandoff && !s->zhttpAddress.isEmpty())
-				{
-					ZhttpRequestPacket zreq;
-					zreq.type = ZhttpRequestPacket::KeepAlive;
-					zhttp_out_write(s, zreq);
-				}
-			}
-		}
-
-		{
-			QHashIterator<Rid, Session*> it(sessionsByZwsRid);
-			while(it.hasNext())
-			{
-				it.next();
-				Session *s = it.value();
-
-				if(!s->inHandoff && !s->zhttpAddress.isEmpty())
-				{
-					ZhttpRequestPacket zreq;
-					zreq.type = ZhttpRequestPacket::KeepAlive;
-					zhttp_out_write(s, zreq);
-				}
-			}
-		}
-	}
-
-	void m2KeepAlive_timeout()
-	{
-		QHashIterator<Rid, Session*> it(sessionsByM2Rid);
-		while(it.hasNext())
-		{
-			it.next();
-			Session *s = it.value();
-
-			QVariantHash args;
-			args["keep-alive"] = true;
-			m2_writeCtl(s->conn, args);
-		}
+		refreshM2Connections(now);
+		refreshSessions(now);
+		expireSessions(now);
+		cancelSessions();
 	}
 
 	void reload()
