@@ -46,6 +46,9 @@ static QVariant acceptDataToVariant(const AcceptData &adata)
 			if(r.debug)
 				vrequest["debug"] = true;
 
+			if(r.isRetry)
+				vrequest["is-retry"] = true;
+
 			if(r.autoCrossOrigin)
 				vrequest["auto-cross-origin"] = true;
 
@@ -137,6 +140,15 @@ static QVariant acceptDataToVariant(const AcceptData &adata)
 	if(!adata.channelPrefix.isEmpty())
 		obj["channel-prefix"] = adata.channelPrefix;
 
+	if(!adata.sigIss.isEmpty())
+		obj["sig-iss"] = adata.sigIss;
+
+	if(!adata.sigKey.isEmpty())
+		obj["sig-key"] = adata.sigKey;
+
+	if(adata.trusted)
+		obj["trusted"] = true;
+
 	if(adata.useSession)
 		obj["use-session"] = true;
 
