@@ -260,6 +260,12 @@ public:
 		QString handler_ws_control_out_spec = settings.value("proxy/handler_ws_control_out_spec").toString();
 		QString stats_spec = settings.value("proxy/stats_spec").toString();
 		QString command_spec = settings.value("proxy/command_spec").toString();
+		QStringList intreq_in_specs = settings.value("proxy/intreq_in_specs").toStringList();
+		trimlist(&intreq_in_specs);
+		QStringList intreq_in_stream_specs = settings.value("proxy/intreq_in_stream_specs").toStringList();
+		trimlist(&intreq_in_stream_specs);
+		QStringList intreq_out_specs = settings.value("proxy/intreq_out_specs").toStringList();
+		trimlist(&intreq_out_specs);
 		bool ok;
 		int ipcFileMode = settings.value("proxy/ipc_file_mode", -1).toString().toInt(&ok, 8);
 		int maxWorkers = settings.value("proxy/max_open_requests", -1).toInt();
@@ -275,7 +281,7 @@ public:
 		QByteArray sigKey = parse_key(settings.value("proxy/sig_key").toString());
 		QByteArray upstreamKey = parse_key(settings.value("proxy/upstream_key").toString());
 		QString sockJsUrl = settings.value("proxy/sockjs_url").toString();
-		bool updatesCheck = settings.value("proxy/updates_check", true).toBool();
+		QString updatesCheck = settings.value("proxy/updates_check").toString();
 		QString organizationName = settings.value("proxy/organization_name").toString();
 
 		QList<QByteArray> origHeadersNeedMark;
@@ -294,6 +300,9 @@ public:
 			return;
 		}
 
+		if(updatesCheck == "true")
+			updatesCheck = "check";
+
 		Engine::Configuration config;
 		config.appVersion = VERSION;
 		config.clientId = "pushpin-proxy_" + QByteArray::number(QCoreApplication::applicationPid());
@@ -310,6 +319,9 @@ public:
 		config.wsControlOutSpec = handler_ws_control_out_spec;
 		config.statsSpec = stats_spec;
 		config.commandSpec = command_spec;
+		config.intServerInSpecs = intreq_in_specs;
+		config.intServerInStreamSpecs = intreq_in_stream_specs;
+		config.intServerOutSpecs = intreq_out_specs;
 		config.ipcFileMode = ipcFileMode;
 		config.maxWorkers = maxWorkers;
 		if(!args.routeLines.isEmpty())
