@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2013 Fanout, Inc.
+ * Copyright (C) 2016 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -33,7 +33,7 @@ public:
 	void start();
 
 signals:
-	void quit();
+	void quit(int returnCode = 0);
 
 private:
 	class Private;
