@@ -30,6 +30,7 @@
 #include <QTextStream>
 #include <QFileSystemWatcher>
 #include "log.h"
+#include "routesfile.h"
 
 static QByteArray parse_key(const QString &in)
 {
@@ -39,73 +40,6 @@ static QByteArray parse_key(const QString &in)
 		return in.toUtf8();
 }
 
-// items are of the format: {value}(,propname=propval,...)
-static bool parseItem(const QString &item, QString *_value, QHash<QString, QString> *_props, QString *errmsg)
-{
-	// read value
-	int at = item.indexOf(',');
-	QString value;
-	if(at != -1)
-		value = item.mid(0, at);
-	else
-		value = item;
-
-	if(value.isEmpty())
-	{
-		*errmsg = "empty item value";
-		return false;
-	}
-
-	// read props
-	QHash<QString, QString> props;
-	int start = at + 1;
-	bool done = false;
-	while(!done)
-	{
-		at = item.indexOf(',', start);
-
-		QString attrib;
-		if(at != -1)
-		{
-			attrib = item.mid(start, at - start);
-			start = at + 1;
-		}
-		else
-		{
-			attrib = item.mid(start);
-			done = true;
-		}
-
-		at = attrib.indexOf('=');
-		QString var, val;
-		if(at != -1)
-		{
-			var = attrib.mid(0, at);
-			val = attrib.mid(at + 1);
-		}
-		else
-			var = attrib;
-
-		if(var.isEmpty())
-		{
-			*errmsg = "empty property name";
-			return false;
-		}
-
-		if(props.contains(var))
-		{
-			*errmsg = "duplicate property: " + var;
-			return false;
-		}
-
-		props[var] = val;
-	}
-
-	*_value = value;
-	*_props = props;
-	return true;
-}
-
 class DomainMap::Worker : public QObject
 {
 	Q_OBJECT
@@ -134,6 +68,7 @@ public:
 		bool session;
 		QByteArray sockJsPath;
 		QByteArray sockJsAsPath;
+		HttpHeaders headers;
 		QList<Target> targets;
 
 		Rule() :
@@ -208,6 +143,7 @@ public:
 			e.session = session;
 			e.sockJsPath = sockJsPath;
 			e.sockJsAsPath = sockJsAsPath;
+			e.headers = headers;
 			e.targets = targets;
 			return e;
 		}
@@ -241,15 +177,6 @@ public:
 		{
 			QString line = ts.readLine();
 
-			// strip comments
-			int at = line.indexOf('#');
-			if(at != -1)
-				line.truncate(at);
-
-			line = line.trimmed();
-			if(line.isEmpty())
-				continue;
-
 			Rule r;
 			if(!parseRouteLine(line, fileName, lineNum, &r))
 			{
@@ -355,24 +282,32 @@ public slots:
 private:
 	static bool parseRouteLine(const QString &line, const QString &fileName, int lineNum, Rule *rule)
 	{
-		QStringList parts = line.split(' ', QString::SkipEmptyParts);
-		if(parts.count() < 2)
-		{
-			log_warning("%s:%d: must specify rule and at least one target", qPrintable(fileName), lineNum);
-			return false;
-		}
-
-		QString val;
-		QHash<QString, QString> props;
+		bool ok;
 		QString errmsg;
-		if(!parseItem(parts[0], &val, &props, &errmsg))
+		QList<RoutesFile::RouteSection> sections = RoutesFile::parseLine(line, &ok, &errmsg);
+		if(!ok)
 		{
 			log_warning("%s:%d: %s", qPrintable(fileName), lineNum, qPrintable(errmsg));
 			return false;
 		}
 
+		if(sections.isEmpty())
+		{
+			// nothing. could happen if line is blank or commented out
+			return false;
+		}
+
+		if(sections.count() < 2)
+		{
+			log_warning("%s:%d: must specify rule and at least one target", qPrintable(fileName), lineNum);
+			return false;
+		}
+
+		QString val = sections[0].value;
+		QHash<QString, QString> props = sections[0].props;
+
 		if(val == "*")
-			val = QString();
+			val.clear();
 
 		Rule r;
 		r.domain = val;
@@ -506,15 +441,38 @@ private:
 		if(props.contains("sockjs_as_path"))
 			r.sockJsAsPath = props.value("sockjs_as_path").toUtf8();
 
-		bool ok = true;
-		for(int n = 1; n < parts.count(); ++n)
+		if(props.contains("header"))
 		{
-			if(!parseItem(parts[n], &val, &props, &errmsg))
+			foreach(const QString &s, props.values("header"))
 			{
-				log_warning("%s:%d: %s", qPrintable(fileName), lineNum, qPrintable(errmsg));
-				ok = false;
-				break;
+				int at = s.indexOf(':');
+				if(at < 1)
+				{
+					log_warning("%s:%d: header must use format 'name:value'", qPrintable(fileName), lineNum);
+					return false;
+				}
+
+				QByteArray name = s.mid(0, at).toUtf8();
+				QByteArray value = s.mid(at + 1).toUtf8();
+
+				// trim left side of value
+				int n = 0;
+				while(n < value.length() && value[n] == ' ')
+				{
+					++n;
+				}
+				if(n > 0)
+					value = value.mid(n);
+
+				r.headers += HttpHeader(name, value);
 			}
+		}
+
+		ok = true;
+		for(int n = 1; n < sections.count(); ++n)
+		{
+			QString val = sections[n].value;
+			QHash<QString, QString> props = sections[n].props;
 
 			Target target;
 
@@ -556,7 +514,7 @@ private:
 					break;
 				}
 
-				target.connectHost = parts[n].mid(0, at);
+				target.connectHost = val.mid(0, at);
 				target.connectPort = port;
 			}
 
@@ -568,6 +526,9 @@ private:
 			else
 				target.trusted = true;
 
+			if(props.contains("trust_connect_host"))
+				target.trustConnectHost = true;
+
 			if(props.contains("insecure"))
 				target.insecure = true;
 
