@@ -23,10 +23,11 @@
 #include "qzmqsocket.h"
 #include "qzmqvalve.h"
 #include "tnetstring.h"
+#include "packet/httpresponsedata.h"
 #include "packet/retryrequestpacket.h"
+#include "packet/statspacket.h"
 #include "log.h"
 #include "inspectdata.h"
-#include "acceptdata.h"
 #include "zhttpmanager.h"
 #include "zhttprequest.h"
 #include "zwebsocket.h"
@@ -49,11 +50,6 @@
 
 #define DEFAULT_HWM 1000
 
-static QByteArray ridToString(const QPair<QByteArray, QByteArray> &rid)
-{
-	return rid.first + ':' + rid.second;
-}
-
 class Engine::Private : public QObject
 {
 	Q_OBJECT
@@ -88,6 +84,7 @@ public:
 	bool destroying;
 	Configuration config;
 	ZhttpManager *zhttpIn;
+	ZhttpManager *intZhttpIn;
 	ZRoutes *zroutes;
 	ZrpcManager *inspect;
 	WsControlManager *wsControl;
@@ -104,12 +101,14 @@ public:
 	QHash<WsProxySession*, WsProxyItem*> wsProxyItemsBySession;
 	SockJsManager *sockJsManager;
 	ConnectionManager connectionManager;
+	Updater *updater;
 
 	Private(Engine *_q) :
 		QObject(_q),
 		q(_q),
 		destroying(false),
 		zhttpIn(0),
+		intZhttpIn(0),
 		zroutes(0),
 		inspect(0),
 		wsControl(0),
@@ -120,7 +119,8 @@ public:
 		accept(0),
 		handler_retry_in_sock(0),
 		handler_retry_in_valve(0),
-		sockJsManager(0)
+		sockJsManager(0),
+		updater(0)
 	{
 	}
 
@@ -128,6 +128,10 @@ public:
 	{
 		destroying = true;
 
+		// need to delete all objects that may have outgoing connections before zroutes
+
+		delete updater;
+
 		QHashIterator<ProxySession*, ProxyItem*> it(proxyItemsBySession);
 		while(it.hasNext())
 		{
@@ -184,6 +188,19 @@ public:
 		zhttpIn->setServerInStreamSpecs(config.serverInStreamSpecs);
 		zhttpIn->setServerOutSpecs(config.serverOutSpecs);
 
+		if(!config.intServerInSpecs.isEmpty() && !config.intServerInStreamSpecs.isEmpty() && !config.intServerOutSpecs.isEmpty())
+		{
+			intZhttpIn = new ZhttpManager(this);
+			intZhttpIn->setBind(true);
+			intZhttpIn->setIpcFileMode(config.ipcFileMode);
+			connect(intZhttpIn, &ZhttpManager::requestReady, this, &Private::intZhttpIn_requestReady);
+
+			intZhttpIn->setInstanceId(config.clientId);
+			intZhttpIn->setServerInSpecs(config.intServerInSpecs);
+			intZhttpIn->setServerInStreamSpecs(config.intServerInStreamSpecs);
+			intZhttpIn->setServerOutSpecs(config.intServerOutSpecs);
+		}
+
 		zroutes = new ZRoutes(this);
 		zroutes->setInstanceId(config.clientId);
 		zroutes->setDefaultOutSpecs(config.clientOutSpecs);
@@ -291,9 +308,9 @@ public:
 			}
 		}
 
-		if(!config.appVersion.isEmpty() && config.updatesCheck)
+		if(!config.appVersion.isEmpty() && (config.updatesCheck == "check" || config.updatesCheck == "report"))
 		{
-			new Updater(config.appVersion, config.organizationName, zroutes->defaultManager(), this);
+			updater = new Updater(config.updatesCheck == "report" ? Updater::ReportMode : Updater::CheckMode, config.appVersion, config.organizationName, zroutes->defaultManager(), this);
 		}
 
 		// init zroutes
@@ -307,7 +324,7 @@ public:
 		domainMap->reload();
 	}
 
-	void doProxy(RequestSession *rs, const InspectData *idata = 0, bool isRetry = false)
+	void doProxy(RequestSession *rs, const InspectData *idata = 0)
 	{
 		DomainMap::Entry route = rs->route();
 
@@ -365,12 +382,6 @@ public:
 		rs->disconnect(this);
 
 		ps->add(rs);
-
-		if(stats)
-		{
-			stats->addConnection(ridToString(rs->rid()), route.id, StatsManager::Http, rs->peerAddress(), rs->isHttps(), isRetry);
-			stats->addActivity(route.id);
-		}
 	}
 
 	void doProxySocket(WebSocket *sock, const DomainMap::Entry &route)
@@ -380,6 +391,8 @@ public:
 		WsProxySession *ps = new WsProxySession(zroutes, &connectionManager, stats, wsControl, this);
 		connect(ps, &WsProxySession::finishedByPassthrough, this, &Private::wsps_finishedByPassthrough);
 
+		connectionManager.setProxyForConnection(sock, ps);
+
 		ps->setDefaultSigKey(config.sigIss, config.sigKey);
 		ps->setDefaultUpstreamKey(config.upstreamKey);
 		ps->setAcceptXForwardedProtocol(config.acceptXForwardedProtocol);
@@ -416,7 +429,9 @@ public:
 
 	bool isXForwardedProtocolTls(const HttpHeaders &headers)
 	{
-		QByteArray xfp = headers.get("X-Forwarded-Protocol");
+		QByteArray xfp = headers.get("X-Forwarded-Proto");
+		if(xfp.isEmpty())
+			xfp = headers.get("X-Forwarded-Protocol");
 		return (!xfp.isEmpty() && (xfp == "https" || xfp == "wss"));
 	}
 
@@ -425,23 +440,83 @@ public:
 		if(!canTake())
 			return;
 
+		// prioritize external requests over internal requests
+
 		ZhttpRequest *req = zhttpIn->takeNextRequest();
 		if(!req)
-			return;
+		{
+			if(intZhttpIn)
+				req = intZhttpIn->takeNextRequest();
 
-		if(config.acceptXForwardedProtocol && isXForwardedProtocolTls(req->requestHeaders()))
-			req->setIsTls(true);
+			if(!req)
+				return;
+		}
+
+		bool lookupRoute = true;
+		bool autoShare = false;
+
+		QVariant passthroughData = req->passthroughData();
+		if(passthroughData.isValid())
+		{
+			if(passthroughData.type() == QVariant::Hash)
+			{
+				QVariantHash data = passthroughData.toHash();
+
+				if(data.contains("route"))
+					lookupRoute = data["route"].toBool();
+
+				if(data.contains("auto-share"))
+					autoShare = data["auto-share"].toBool();
+			}
+		}
+		else
+		{
+			if(config.acceptXForwardedProtocol && isXForwardedProtocolTls(req->requestHeaders()))
+				req->setIsTls(true);
+		}
+
+		RequestSession *rs;
+		if(lookupRoute)
+		{
+			rs = new RequestSession(domainMap, sockJsManager, inspect, inspectChecker, accept, stats, this);
+			rs->setDebugEnabled(config.debug);
+			rs->setAutoCrossOrigin(config.autoCrossOrigin);
+			rs->setPrefetchSize(config.inspectPrefetch);
+		}
+		else
+		{
+			assert(passthroughData.isValid());
+
+			const QVariantHash data = passthroughData.toHash();
+
+			// make a direct route, no domainmap lookup
+			DomainMap::Entry route;
+			DomainMap::Target target;
+			QUrl uri = req->requestUri();
+			bool isHttps = (uri.scheme() == "https");
+			target.connectHost = uri.host();
+			target.connectPort = uri.port(isHttps ? 443 : 80);
+			target.ssl = isHttps;
+			if(passthroughData.type() == QVariant::Hash)
+			{
+
+				route.sigIss = data["sig-iss"].toByteArray();
+				route.sigKey = data["sig-key"].toByteArray();
+				target.trusted = data["trusted"].toBool();
+			}
+			route.targets += target;
+
+			rs = new RequestSession(stats, this);
+			rs->setRoute(route);
+		}
+
+		rs->setAutoShare(autoShare);
 
-		RequestSession *rs = new RequestSession(domainMap, sockJsManager, inspect, inspectChecker, accept, this);
 		connect(rs, &RequestSession::inspected, this, &Private::rs_inspected);
 		connect(rs, &RequestSession::inspectError, this, &Private::rs_inspectError);
 		connect(rs, &RequestSession::finished, this, &Private::rs_finished);
 		connect(rs, &RequestSession::finishedByAccept, this, &Private::rs_finishedByAccept);
 
-		rs->setDebugEnabled(config.debug);
-		rs->setAutoCrossOrigin(config.autoCrossOrigin);
-		rs->setPrefetchSize(config.inspectPrefetch);
-
 		requestSessions += rs;
 
 		rs->start(req);
@@ -520,7 +595,7 @@ public:
 
 		if(accepted)
 		{
-			msg += " hold";
+			msg += " accept";
 		}
 		else if(resp.code != -1)
 		{
@@ -550,6 +625,11 @@ private slots:
 		tryTakeNext();
 	}
 
+	void intZhttpIn_requestReady()
+	{
+		tryTakeNext();
+	}
+
 	void rs_inspected(const InspectData &idata)
 	{
 		RequestSession *rs = (RequestSession *)sender();
@@ -575,9 +655,6 @@ private slots:
 
 		logFinished(rs);
 
-		if(stats)
-			stats->removeConnection(ridToString(rs->rid()), false);
-
 		requestSessions.remove(rs);
 		delete rs;
 
@@ -590,16 +667,6 @@ private slots:
 
 		logFinished(rs, true);
 
-		if(stats)
-		{
-			// add connection so that it becomes lingerable
-			stats->addConnection(ridToString(rs->rid()), QByteArray(), StatsManager::Http, rs->peerAddress(), rs->isHttps(), false);
-			stats->addActivity(QByteArray());
-
-			// immediately remove since we're accepting
-			stats->removeConnection(ridToString(rs->rid()), true);
-		}
-
 		requestSessions.remove(rs);
 		delete rs;
 
@@ -641,8 +708,7 @@ private slots:
 	{
 		requestSessions.remove(rs);
 
-		if(stats)
-			stats->removeConnection(ridToString(rs->rid()), accept);
+		rs->setAccepted(accept);
 
 		tryTakeNext();
 	}
@@ -718,7 +784,7 @@ private slots:
 
 			ZhttpRequest *zhttpRequest = zhttpIn->createRequestFromState(ss);
 
-			RequestSession *rs = new RequestSession(domainMap, sockJsManager, inspect, inspectChecker, accept, this);
+			RequestSession *rs = new RequestSession(domainMap, sockJsManager, inspect, inspectChecker, accept, stats, this);
 			requestSessions += rs;
 
 			// note: if the routing table was changed, there's a chance the request
@@ -726,7 +792,7 @@ private slots:
 			//   stats processors tracking route+connection mappings.
 			rs->startRetry(zhttpRequest, req.debug, req.autoCrossOrigin, req.jsonpCallback, req.jsonpExtendedResponse);
 
-			doProxy(rs, p.haveInspectInfo ? &idata : 0, true);
+			doProxy(rs, p.haveInspectInfo ? &idata : 0);
 		}
 	}
 
@@ -780,6 +846,72 @@ private slots:
 
 			req->respond(out);
 		}
+		else if(req->method() == "refresh")
+		{
+			QVariantHash args = req->args();
+			if(!args.contains("cid") || args["cid"].type() != QVariant::ByteArray)
+			{
+				req->respondError("bad-format");
+				delete req;
+				return;
+			}
+
+			QByteArray cid = args["cid"].toByteArray();
+
+			WsProxySession *ps = connectionManager.getProxyForConnection(cid);
+			if(!ps)
+			{
+				req->respondError("item-not-found");
+				delete req;
+				return;
+			}
+
+			WebSocketOverHttp *woh = qobject_cast<WebSocketOverHttp*>(ps->outSocket());
+			if(woh)
+				woh->refresh();
+
+			req->respond();
+		}
+		else if(req->method() == "report")
+		{
+			QVariantHash args = req->args();
+			if(!args.contains("stats") || args["stats"].type() != QVariant::Hash)
+			{
+				req->respondError("bad-format");
+				delete req;
+				return;
+			}
+
+			QVariant data = args["stats"];
+
+			StatsPacket p;
+			if(!p.fromVariant("report", data))
+			{
+				req->respondError("bad-format");
+				delete req;
+				return;
+			}
+
+			if(!updater)
+			{
+				req->respondError("service-unavailable");
+				delete req;
+				return;
+			}
+
+			Updater::Report report;
+			report.connectionsMax = p.connectionsMax;
+			report.connectionsMinutes = p.connectionsMinutes;
+			report.messagesReceived = p.messagesReceived;
+			report.messagesSent = p.messagesSent;
+
+			// fanout cloud style ops calculation
+			report.ops = p.connectionsMinutes + p.messagesReceived + p.messagesSent - p.httpResponseMessagesSent;
+
+			updater->setReport(report);
+
+			req->respond();
+		}
 		else
 		{
 			req->respondError("method-not-found");
