@@ -47,6 +47,9 @@ public:
 		QString wsControlOutSpec;
 		QString statsSpec;
 		QString commandSpec;
+		QStringList intServerInSpecs;
+		QStringList intServerInStreamSpecs;
+		QStringList intServerOutSpecs;
 		int ipcFileMode;
 		int maxWorkers;
 		int inspectTimeout;
@@ -64,7 +67,7 @@ public:
 		QByteArray sigKey;
 		QByteArray upstreamKey;
 		QString sockJsUrl;
-		bool updatesCheck;
+		QString updatesCheck;
 		QString organizationName;
 
 		Configuration() :
@@ -76,7 +79,7 @@ public:
 			autoCrossOrigin(false),
 			acceptXForwardedProtocol(false),
 			useXForwardedProtocol(false),
-			updatesCheck(true)
+			updatesCheck("check")
 		{
 		}
 	};
