@@ -49,7 +49,7 @@ static bool validate_token(const QByteArray &token, const QByteArray &key)
 
 namespace ProxyUtil {
 
-bool manipulateRequestHeaders(const char *logprefix, void *object, HttpRequestData *requestData, const QByteArray &defaultUpstreamKey, const DomainMap::Entry &entry, const QByteArray &sigIss, const QByteArray &sigKey, bool acceptXForwardedProtocol, bool useXForwardedProtocol, const XffRule &xffTrustedRule, const XffRule &xffRule, const QList<QByteArray> &origHeadersNeedMark, const QHostAddress &peerAddress, const InspectData &idata)
+bool manipulateRequestHeaders(const char *logprefix, void *object, HttpRequestData *requestData, const QByteArray &defaultUpstreamKey, const DomainMap::Entry &entry, const QByteArray &sigIss, const QByteArray &sigKey, bool acceptXForwardedProtocol, bool useXForwardedProtocol, const XffRule &xffTrustedRule, const XffRule &xffRule, const QList<QByteArray> &origHeadersNeedMark, const QHostAddress &peerAddress, const InspectData &idata, bool stripHeaders)
 {
 	// check if the request is coming from a grip proxy already
 	bool trustedClient = false;
@@ -136,21 +136,18 @@ bool manipulateRequestHeaders(const char *logprefix, void *object, HttpRequestDa
 	requestData->headers.removeAll("Transfer-Encoding");
 	requestData->headers.removeAll("Expect");
 
-	// rewrite the Host header to match the hostname of the destination URL.
-	//   in practice, the only time the value should ever be different is
-	//   if the original Host header had a port specified
-	requestData->headers.removeAll("Host");
-	requestData->headers += HttpHeader("Host", requestData->uri.host().toUtf8());
-
 	if(!trustedClient)
 	{
-		// remove all Grip- headers
-		for(int n = 0; n < requestData->headers.count(); ++n)
+		if(stripHeaders)
 		{
-			if(qstrnicmp(requestData->headers[n].first.data(), "Grip-", 5) == 0)
+			// remove all Grip- headers
+			for(int n = 0; n < requestData->headers.count(); ++n)
 			{
-				requestData->headers.removeAt(n);
-				--n; // adjust position
+				if(qstrnicmp(requestData->headers[n].first.data(), "Grip-", 5) == 0)
+				{
+					requestData->headers.removeAt(n);
+					--n; // adjust position
+				}
 			}
 		}
 
@@ -180,13 +177,25 @@ bool manipulateRequestHeaders(const char *logprefix, void *object, HttpRequestDa
 	}
 
 	if(acceptXForwardedProtocol || useXForwardedProtocol)
+	{
+		requestData->headers.removeAll("X-Forwarded-Proto");
+
+		// TODO: deprecate
 		requestData->headers.removeAll("X-Forwarded-Protocol");
+	}
 
 	if(useXForwardedProtocol)
 	{
 		QString scheme = requestData->uri.scheme();
 		if(scheme == "https" || scheme == "wss")
-			requestData->headers += HttpHeader("X-Forwarded-Protocol", scheme.toUtf8());
+		{
+			QByteArray schemeVal = scheme.toUtf8();
+
+			requestData->headers += HttpHeader("X-Forwarded-Proto", schemeVal);
+
+			// TODO: deprecate
+			requestData->headers += HttpHeader("X-Forwarded-Protocol", schemeVal);
+		}
 	}
 
 	const XffRule *xr;
@@ -206,4 +215,32 @@ bool manipulateRequestHeaders(const char *logprefix, void *object, HttpRequestDa
 	return trustedClient;
 }
 
+void applyHost(QUrl *url, const QString &host)
+{
+	int at = host.indexOf(':');
+	if(at != -1)
+	{
+		url->setHost(host.mid(0, at));
+		url->setPort(host.mid(at + 1).toInt());
+	}
+	else
+	{
+		url->setHost(host);
+		url->setPort(-1);
+	}
+}
+
+void applyHostHeader(HttpHeaders *headers, const QUrl &uri)
+{
+	QByteArray hostHeader = uri.host().toUtf8();
+	if(uri.port() != -1)
+		hostHeader += ':' + QByteArray::number(uri.port());
+
+	if(headers->get("Host") != hostHeader)
+	{
+		headers->removeAll("Host");
+		headers->append(HttpHeader("Host", hostHeader));
+	}
+}
+
 }
