@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013-2015 Fanout, Inc.
+ * Copyright (C) 2013-2016 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -31,6 +31,9 @@
 #include "packet/httpresponsedata.h"
 #include "engine.h"
 
+// NOTE: based on proxysession hardcoded max
+#define PROXY_MAX_ACCEPT_RESPONSE_BODY 100000
+
 class Wrapper : public QObject
 {
 	Q_OBJECT
@@ -58,6 +61,7 @@ public:
 	bool inspected;
 	QByteArray sharingKey;
 	QByteArray in;
+	HttpHeaders acceptHeaders;
 	QByteArray acceptIn;
 	bool retried;
 	bool finished;
@@ -146,6 +150,7 @@ public:
 		inspected = false;
 		sharingKey.clear();
 		in.clear();
+		acceptHeaders.clear();
 		acceptIn.clear();
 		retried = false;
 		finished = false;
@@ -165,16 +170,16 @@ private slots:
 		zresp.fromVariant(v);
 		if(zresp.type == ZhttpResponsePacket::Data)
 		{
-			if(!responses.contains(zresp.id))
+			if(!responses.contains(zresp.ids.first().id))
 			{
 				HttpResponseData rd;
 				rd.code = zresp.code;
 				rd.reason = zresp.reason;
 				rd.headers = zresp.headers;
-				responses[zresp.id] = rd;
+				responses[zresp.ids.first().id] = rd;
 			}
 
-			responses[zresp.id].body += zresp.body;
+			responses[zresp.ids.first().id].body += zresp.body;
 			in += zresp.body;
 
 			if(!zresp.more)
@@ -187,8 +192,7 @@ private slots:
 		{
 			ZhttpRequestPacket zreq;
 			zreq.from = "test-client";
-			zreq.id = zresp.id;
-			zreq.seq = 1;
+			zreq.ids += ZhttpRequestPacket::Id(zresp.ids.first().id, 1);
 			zreq.type = ZhttpRequestPacket::HandoffProceed;
 			QByteArray buf = 'T' + TnetString::fromVariant(zreq.toVariant());
 			log_debug("writing: %s", buf.data());
@@ -239,8 +243,7 @@ private slots:
 			{
 				ZhttpResponsePacket zresp;
 				zresp.from = "test-server";
-				zresp.id = zreq.id;
-				zresp.seq = serverOutSeq++;
+				zresp.ids += ZhttpResponsePacket::Id(zreq.ids.first().id, serverOutSeq++);
 				zresp.type = ZhttpResponsePacket::Credit;
 				zresp.credits = 200000;
 				QByteArray buf = zreq.from + " T" + TnetString::fromVariant(zresp.toVariant());
@@ -252,37 +255,103 @@ private slots:
 
 		ZhttpResponsePacket zresp;
 		zresp.from = "test-server";
-		zresp.id = zreq.id;
-		zresp.seq = serverOutSeq++;
+		zresp.ids += ZhttpResponsePacket::Id(zreq.ids.first().id, serverOutSeq++);
 		zresp.code = 200;
 		zresp.reason = "OK";
 
 		QByteArray encPath = zreq.uri.path(QUrl::FullyEncoded).toUtf8();
 
-		if(!retried && zreq.uri.query(QUrl::FullyEncoded).contains("wait=true"))
+		QUrlQuery query(zreq.uri.query());
+		QString hold = query.queryItemValue("hold");
+		bool bodyInstruct = (query.queryItemValue("body-instruct") == "true");
+		bool large = (query.queryItemValue("large") == "true");
+
+		if(!retried && (hold == "response" || hold == "stream"))
 		{
 			if(encPath == "/path2")
 			{
-				zresp.body = "{ \"hold\": { \"mode\": \"response\", \"channels\": [ { \"name\": \"test-channel\", \"prev-id\": \"1\" } ] } }";
-				zresp.headers += HttpHeader("Content-Type", "application/grip-instruct");
+				if(bodyInstruct)
+				{
+					zresp.headers += HttpHeader("Content-Type", "application/grip-instruct");
+					zresp.body = "{ \"hold\": { \"mode\": \"response\", \"channels\": [ { \"name\": \"test-channel\", \"prev-id\": \"1\" } ] } }";
+				}
+				else
+				{
+					zresp.headers += HttpHeader("Grip-Hold", "response");
+					zresp.headers += HttpHeader("Grip-Channel", "test-channel; prev-id=1");
+				}
 			}
 			else
 			{
-				zresp.body = "{ \"hold\": { \"mode\": \"response\", \"channels\": [ { \"name\": \"test-channel\" } ] } }";
-				zresp.headers += HttpHeader("Content-Type", "application/grip-instruct");
+				if(bodyInstruct)
+				{
+					zresp.headers += HttpHeader("Content-Type", "application/grip-instruct");
+					zresp.body = "{ \"hold\": { \"mode\": \"response\", \"channels\": [ { \"name\": \"test-channel\" } ] } }";
+				}
+				else
+				{
+					if(hold == "stream")
+					{
+						zresp.headers += HttpHeader("Grip-Hold", "stream");
+						zresp.headers += HttpHeader("Grip-Channel", "test-channel");
+						if(large)
+							zresp.body = QByteArray(PROXY_MAX_ACCEPT_RESPONSE_BODY + 10000, 'a') + '\n';
+						else
+							zresp.body = "stream open\n";
+					}
+					else
+					{
+						zresp.headers += HttpHeader("Grip-Hold", "response");
+						zresp.headers += HttpHeader("Grip-Channel", "test-channel");
+					}
+				}
 			}
 		}
 		else
 		{
 			if(encPath.startsWith("/jsonp"))
 			{
-				zresp.body = "{\"hello\": \"world\"}";
 				zresp.headers += HttpHeader("Content-Type", "application/json");
+				zresp.body = "{\"hello\": \"world\"}";
+			}
+			else if(encPath == "/path3")
+			{
+				zresp.headers += HttpHeader("Content-Type", "text/plain");
+				zresp.body = "next page";
 			}
 			else
 			{
-				zresp.body = "hello world";
-				zresp.headers += HttpHeader("Content-Type", "text/plain");
+				if(hold == "none")
+				{
+					if(bodyInstruct)
+					{
+						zresp.headers += HttpHeader("Content-Type", "application/grip-instruct");
+						zresp.body = "{ \"response\": { \"body\": \"hello world\" } }";
+					}
+					else
+					{
+						zresp.headers += HttpHeader("Content-Type", "text/plain");
+						if(large)
+						{
+							// Grip-Link required to trigger accept after
+							//   sending large response. note that the link
+							//   won't be followed in this test since that's
+							//   not a proxy issue
+							zresp.headers += HttpHeader("Grip-Link", "</path3>; rel=next");
+							zresp.body = QByteArray(PROXY_MAX_ACCEPT_RESPONSE_BODY + 10000, 'a') + '\n';
+						}
+						else
+						{
+							zresp.headers += HttpHeader("Grip-Foo", "bar"); // something to trigger accept
+							zresp.body = "hello world";
+						}
+					}
+				}
+				else
+				{
+					zresp.headers += HttpHeader("Content-Type", "text/plain");
+					zresp.body = "hello world";
+				}
 			}
 		}
 		zresp.headers += HttpHeader("Content-Length", QByteArray::number(zresp.body.size()));
@@ -325,38 +394,67 @@ private slots:
 		log_debug("accept: %s", qPrintable(TnetString::variantToString(v, -1)));
 
 		QVariantHash vreq = v.toHash();
+		QVariantHash vaccept = vreq["args"].toHash();
+		QVariantHash vresponse = vaccept["response"].toHash();
+
+		acceptHeaders.clear();
+		foreach(const QVariant &vheader, vresponse["headers"].toList())
+		{
+			QVariantList h = vheader.toList();
+			acceptHeaders += HttpHeader(h[0].toByteArray(), h[1].toByteArray());
+		}
+
+		acceptIn = vresponse["body"].toByteArray();
+
+		QVariantMap jsonInstruct;
+		QByteArray hold;
+
+		if(acceptHeaders.get("Content-Type") == "application/grip-instruct")
+		{
+			QJsonParseError e;
+			QJsonDocument doc = QJsonDocument::fromJson(acceptIn, &e);
+			QVERIFY(e.error == QJsonParseError::NoError);
+			QVERIFY(doc.isObject());
+
+			jsonInstruct = doc.object().toVariantMap();
+
+			if(jsonInstruct.contains("hold"))
+				hold = jsonInstruct["hold"].toMap().value("mode").toString().toUtf8();
+		}
+		else
+		{
+			hold = acceptHeaders.get("Grip-Hold");
+		}
 
 		QVariantHash vresp;
 		vresp["id"] = vreq["id"];
 		vresp["success"] = true;
 		QVariantHash respValue;
-		respValue["accepted"] = true;
+		if(!hold.isEmpty())
+			respValue["accepted"] = true;
+		else if(!vaccept.value("response-sent").toBool())
+			respValue["response"] = vresponse;
 		vresp["value"] = respValue;
 		handlerAcceptSock->write(message.createReply(QList<QByteArray>() << TnetString::fromVariant(vresp)).toRawMessage());
 
-		QVariantHash vaccept = vreq["args"].toHash();
-		acceptIn = vaccept["response"].toHash()["body"].toByteArray();
-
 		log_debug("instruct: [%s]", acceptIn.data());
 
-		QJsonParseError e;
-		QJsonDocument doc = QJsonDocument::fromJson(acceptIn, &e);
-		QVERIFY(e.error == QJsonParseError::NoError);
-		QVERIFY(doc.isObject());
-
-		QVariantMap instruct = doc.object().toVariantMap();
-
-		if(instruct["hold"].toMap()["channels"].toList()[0].toMap().contains("prev-id"))
+		if(acceptHeaders.get("Content-Type") == "application/grip-instruct")
 		{
-			retried = true;
-			QVariantHash vretry;
-			vretry["requests"] = vaccept["requests"];
-			vretry["request-data"] = vaccept["request-data"];
-			QByteArray buf = TnetString::fromVariant(vretry);
-			log_debug("retrying: %s", buf.data());
-			handlerRetryOutSock->write(QList<QByteArray>() << buf);
+			if(jsonInstruct.contains("hold") && jsonInstruct["hold"].toMap()["channels"].toList()[0].toMap().contains("prev-id"))
+			{
+				retried = true;
+				QVariantHash vretry;
+				vretry["requests"] = vaccept["requests"];
+				vretry["request-data"] = vaccept["request-data"];
+				QByteArray buf = TnetString::fromVariant(vretry);
+				log_debug("retrying: %s", qPrintable(TnetString::variantToString(vretry, -1)));
+				handlerRetryOutSock->write(QList<QByteArray>() << buf);
+				return;
+			}
 		}
-		else
+
+		if(!hold.isEmpty())
 			finished = true;
 	}
 };
@@ -415,8 +513,7 @@ private slots:
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
-		zreq.id = "1";
-		zreq.seq = 0;
+		zreq.ids += ZhttpRequestPacket::Id("1", 0);
 		zreq.uri = "http://example/path";
 		zreq.method = "GET";
 		zreq.stream = true;
@@ -437,8 +534,7 @@ private slots:
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
-		zreq.id = "2";
-		zreq.seq = 0;
+		zreq.ids += ZhttpRequestPacket::Id("2", 0);
 		zreq.uri = "http://example/path";
 		zreq.method = "GET";
 		zreq.stream = true;
@@ -458,8 +554,7 @@ private slots:
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
-		zreq.id = "3";
-		zreq.seq = 0;
+		zreq.ids += ZhttpRequestPacket::Id("3", 0);
 		zreq.uri = "http://example/jsonp?callback=jpcb";
 		zreq.method = "GET";
 		zreq.stream = true;
@@ -491,8 +586,7 @@ private slots:
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
-		zreq.id = "4";
-		zreq.seq = 0;
+		zreq.ids += ZhttpRequestPacket::Id("4", 0);
 		zreq.uri = "http://example/jsonp-basic?bparam={}";
 		zreq.method = "GET";
 		zreq.stream = true;
@@ -512,8 +606,7 @@ private slots:
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
-		zreq.id = "5";
-		zreq.seq = 0;
+		zreq.ids += ZhttpRequestPacket::Id("5", 0);
 		zreq.uri = "http://example/path";
 		zreq.method = "POST";
 		zreq.stream = true;
@@ -532,8 +625,7 @@ private slots:
 		// now finish the request
 		zreq = ZhttpRequestPacket();
 		zreq.from = "test-client";
-		zreq.id = "5";
-		zreq.seq = 1;
+		zreq.ids += ZhttpRequestPacket::Id("5", 1);
 		zreq.type = ZhttpRequestPacket::Data;
 		zreq.body = " world";
 		buf = 'T' + TnetString::fromVariant(zreq.toVariant());
@@ -557,8 +649,7 @@ private slots:
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
-		zreq.id = "6";
-		zreq.seq = 0;
+		zreq.ids += ZhttpRequestPacket::Id("6", 0);
 		zreq.uri = "http://example/path";
 		zreq.method = "POST";
 		zreq.stream = true;
@@ -577,8 +668,7 @@ private slots:
 		// now cancel the request
 		zreq = ZhttpRequestPacket();
 		zreq.from = "test-client";
-		zreq.id = "6";
-		zreq.seq = 1;
+		zreq.ids += ZhttpRequestPacket::Id("6", 1);
 		zreq.type = ZhttpRequestPacket::Cancel;
 		buf = 'T' + TnetString::fromVariant(zreq.toVariant());
 		log_debug("writing: %s", buf.data());
@@ -593,15 +683,59 @@ private slots:
 			QTest::qWait(10);
 	}
 
-	void accept()
+	void acceptResponse()
 	{
 		wrapper->reset();
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
-		zreq.id = "7";
-		zreq.seq = 0;
-		zreq.uri = "http://example/path?wait=true";
+		zreq.ids += ZhttpRequestPacket::Id("7", 0);
+		zreq.uri = "http://example/path?hold=response";
+		zreq.method = "GET";
+		zreq.stream = true;
+		zreq.credits = 200000;
+		QByteArray buf = 'T' + TnetString::fromVariant(zreq.toVariant());
+		log_debug("writing: %s", buf.data());
+		wrapper->zhttpClientOutSock->write(QList<QByteArray>() << buf);
+		while(!wrapper->finished)
+			QTest::qWait(10);
+
+		QCOMPARE(wrapper->acceptHeaders.get("Grip-Hold"), QByteArray("response"));
+		QCOMPARE(wrapper->acceptHeaders.get("Grip-Channel"), QByteArray("test-channel"));
+		QVERIFY(wrapper->acceptIn.isEmpty());
+	}
+
+	void acceptStream()
+	{
+		wrapper->reset();
+
+		ZhttpRequestPacket zreq;
+		zreq.from = "test-client";
+		zreq.ids += ZhttpRequestPacket::Id("8", 0);
+		zreq.uri = "http://example/path?hold=stream";
+		zreq.method = "GET";
+		zreq.stream = true;
+		zreq.credits = 200000;
+		QByteArray buf = 'T' + TnetString::fromVariant(zreq.toVariant());
+		log_debug("writing: %s", buf.data());
+		wrapper->zhttpClientOutSock->write(QList<QByteArray>() << buf);
+		while(!wrapper->finished)
+			QTest::qWait(10);
+
+		QCOMPARE(wrapper->acceptHeaders.get("Grip-Hold"), QByteArray("stream"));
+		QCOMPARE(wrapper->acceptHeaders.get("Grip-Channel"), QByteArray("test-channel"));
+		QCOMPARE(wrapper->acceptIn, QByteArray("stream open\n"));
+		QVERIFY(wrapper->in.isEmpty());
+	}
+
+	void acceptResponseBodyInstruct()
+	{
+		wrapper->reset();
+
+		ZhttpRequestPacket zreq;
+		zreq.from = "test-client";
+		zreq.ids += ZhttpRequestPacket::Id("9", 0);
+		zreq.uri = "http://example/path?hold=response&body-instruct=true";
 		zreq.method = "GET";
 		zreq.stream = true;
 		zreq.credits = 200000;
@@ -614,15 +748,101 @@ private slots:
 		QCOMPARE(wrapper->acceptIn, QByteArray("{ \"hold\": { \"mode\": \"response\", \"channels\": [ { \"name\": \"test-channel\" } ] } }"));
 	}
 
+	void acceptNoHold()
+	{
+		wrapper->reset();
+
+		ZhttpRequestPacket zreq;
+		zreq.from = "test-client";
+		zreq.ids += ZhttpRequestPacket::Id("10", 0);
+		zreq.uri = "http://example/path?hold=none";
+		zreq.method = "GET";
+		zreq.stream = true;
+		zreq.credits = 200000;
+		QByteArray buf = 'T' + TnetString::fromVariant(zreq.toVariant());
+		log_debug("writing: %s", buf.data());
+		wrapper->zhttpClientOutSock->write(QList<QByteArray>() << buf);
+		while(!wrapper->finished)
+			QTest::qWait(10);
+
+		QCOMPARE(wrapper->in, QByteArray("hello world"));
+	}
+
+	void acceptNoHoldBodyInstruct()
+	{
+		wrapper->reset();
+
+		ZhttpRequestPacket zreq;
+		zreq.from = "test-client";
+		zreq.ids += ZhttpRequestPacket::Id("11", 0);
+		zreq.uri = "http://example/path?hold=none&body-instruct=true";
+		zreq.method = "GET";
+		zreq.stream = true;
+		zreq.credits = 200000;
+		QByteArray buf = 'T' + TnetString::fromVariant(zreq.toVariant());
+		log_debug("writing: %s", buf.data());
+		wrapper->zhttpClientOutSock->write(QList<QByteArray>() << buf);
+		while(!wrapper->finished)
+			QTest::qWait(10);
+
+		QCOMPARE(wrapper->acceptIn, QByteArray("{ \"response\": { \"body\": \"hello world\" } }"));
+	}
+
+	void passthroughThenAcceptStream()
+	{
+		wrapper->reset();
+
+		ZhttpRequestPacket zreq;
+		zreq.from = "test-client";
+		zreq.ids += ZhttpRequestPacket::Id("12", 0);
+		zreq.uri = "http://example/path?hold=stream&large=true";
+		zreq.method = "GET";
+		zreq.stream = true;
+		zreq.credits = 200000;
+		QByteArray buf = 'T' + TnetString::fromVariant(zreq.toVariant());
+		log_debug("writing: %s", buf.data());
+		wrapper->zhttpClientOutSock->write(QList<QByteArray>() << buf);
+		while(!wrapper->finished)
+			QTest::qWait(10);
+
+		QCOMPARE(wrapper->in.size(), 110001);
+		QCOMPARE(wrapper->in.mid(wrapper->in.size() - 2), QByteArray("a\n"));
+		QCOMPARE(wrapper->acceptHeaders.get("Grip-Hold"), QByteArray("stream"));
+		QCOMPARE(wrapper->acceptHeaders.get("Grip-Channel"), QByteArray("test-channel"));
+		QVERIFY(wrapper->acceptIn.isEmpty());
+	}
+
+	void passthroughThenAcceptNext()
+	{
+		wrapper->reset();
+
+		ZhttpRequestPacket zreq;
+		zreq.from = "test-client";
+		zreq.ids += ZhttpRequestPacket::Id("13", 0);
+		zreq.uri = "http://example/path?hold=none&large=true";
+		zreq.method = "GET";
+		zreq.stream = true;
+		zreq.credits = 200000;
+		QByteArray buf = 'T' + TnetString::fromVariant(zreq.toVariant());
+		log_debug("writing: %s", buf.data());
+		wrapper->zhttpClientOutSock->write(QList<QByteArray>() << buf);
+		while(!wrapper->finished)
+			QTest::qWait(10);
+
+		QCOMPARE(wrapper->in.size(), 110001);
+		QCOMPARE(wrapper->in.mid(wrapper->in.size() - 2), QByteArray("a\n"));
+		QVERIFY(wrapper->acceptIn.isEmpty());
+		QCOMPARE(wrapper->acceptHeaders.get("Grip-Link"), QByteArray("</path3>; rel=next"));
+	}
+
 	void acceptWithRetry()
 	{
 		wrapper->reset();
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
-		zreq.id = "8";
-		zreq.seq = 0;
-		zreq.uri = "http://example/path2?wait=true";
+		zreq.ids += ZhttpRequestPacket::Id("14", 0);
+		zreq.uri = "http://example/path2?wait=true&body-instruct=true";
 		zreq.method = "GET";
 		zreq.stream = true;
 		zreq.credits = 200000;
@@ -642,7 +862,6 @@ private slots:
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
-		zreq.seq = 0;
 		zreq.uri = "http://example/path";
 		zreq.method = "GET";
 		zreq.stream = true;
@@ -652,12 +871,15 @@ private slots:
 
 		// send two requests
 
-		zreq.id = "9";
+		QByteArray id1 = "15";
+		QByteArray id2 = "16";
+
+		zreq.ids = QList<ZhttpRequestPacket::Id>() << ZhttpRequestPacket::Id(id1, 0);
 		buf = 'T' + TnetString::fromVariant(zreq.toVariant());
 		log_debug("writing: %s", buf.data());
 		wrapper->zhttpClientOutSock->write(QList<QByteArray>() << buf);
 
-		zreq.id = "10";
+		zreq.ids = QList<ZhttpRequestPacket::Id>() << ZhttpRequestPacket::Id(id2, 0);
 		buf = 'T' + TnetString::fromVariant(zreq.toVariant());
 		log_debug("writing: %s", buf.data());
 		wrapper->zhttpClientOutSock->write(QList<QByteArray>() << buf);
@@ -668,8 +890,8 @@ private slots:
 		// there should have only been 1 request to the server
 		QCOMPARE(wrapper->serverReqs, 1);
 
-		QCOMPARE(wrapper->responses["9"].body, QByteArray("hello world"));
-		QCOMPARE(wrapper->responses["10"].body, QByteArray("hello world"));
+		QCOMPARE(wrapper->responses[id1].body, QByteArray("hello world"));
+		QCOMPARE(wrapper->responses[id2].body, QByteArray("hello world"));
 	}
 
 	void passthroughSharedPost()
@@ -679,7 +901,6 @@ private slots:
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
-		zreq.seq = 0;
 		zreq.uri = "http://example/path";
 		zreq.method = "POST";
 		zreq.stream = true;
@@ -691,12 +912,15 @@ private slots:
 
 		// send two requests
 
-		zreq.id = "11";
+		QByteArray id1 = "17";
+		QByteArray id2 = "18";
+
+		zreq.ids = QList<ZhttpRequestPacket::Id>() << ZhttpRequestPacket::Id(id1, 0);
 		buf = 'T' + TnetString::fromVariant(zreq.toVariant());
 		log_debug("writing: %s", buf.data());
 		wrapper->zhttpClientOutSock->write(QList<QByteArray>() << buf);
 
-		zreq.id = "12";
+		zreq.ids = QList<ZhttpRequestPacket::Id>() << ZhttpRequestPacket::Id(id2, 0);
 		buf = 'T' + TnetString::fromVariant(zreq.toVariant());
 		log_debug("writing: %s", buf.data());
 		wrapper->zhttpClientOutSock->write(QList<QByteArray>() << buf);
@@ -709,11 +933,10 @@ private slots:
 
 		zreq = ZhttpRequestPacket();
 		zreq.from = "test-client";
-		zreq.seq = 1;
 		zreq.type = ZhttpRequestPacket::Data;
 		zreq.body = " world";
 
-		zreq.id = "11";
+		zreq.ids = QList<ZhttpRequestPacket::Id>() << ZhttpRequestPacket::Id(id1, 1);
 		buf = 'T' + TnetString::fromVariant(zreq.toVariant());
 		log_debug("writing: %s", buf.data());
 		QList<QByteArray> msg;
@@ -722,7 +945,7 @@ private slots:
 		msg.append(buf);
 		wrapper->zhttpClientOutStreamSock->write(msg);
 
-		zreq.id = "12";
+		zreq.ids = QList<ZhttpRequestPacket::Id>() << ZhttpRequestPacket::Id(id2, 1);
 		buf = 'T' + TnetString::fromVariant(zreq.toVariant());
 		log_debug("writing: %s", buf.data());
 		msg.clear();
@@ -737,8 +960,8 @@ private slots:
 		// there should have only been 1 request to the server
 		QCOMPARE(wrapper->serverReqs, 1);
 
-		QCOMPARE(wrapper->responses["11"].body, QByteArray("hello world"));
-		QCOMPARE(wrapper->responses["12"].body, QByteArray("hello world"));
+		QCOMPARE(wrapper->responses[id1].body, QByteArray("hello world"));
+		QCOMPARE(wrapper->responses[id2].body, QByteArray("hello world"));
 	}
 };
 
