@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 Fanout, Inc.
+ * Copyright (C) 2014-2017 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -172,6 +172,7 @@ public:
 	QString connectHost;
 	int connectPort;
 	bool ignorePolicies;
+	bool trustConnectHost;
 	bool ignoreTlsErrors;
 	State state;
 	QByteArray cid;
@@ -196,6 +197,8 @@ public:
 	bool peerClosing;
 	int peerCloseCode;
 	bool disconnecting;
+	bool disconnectSent;
+	bool updateQueued;
 	QTimer *keepAliveTimer;
 	QTimer *retryTimer;
 	int retries;
@@ -205,6 +208,7 @@ public:
 		q(_q),
 		connectPort(-1),
 		ignorePolicies(false),
+		trustConnectHost(false),
 		ignoreTlsErrors(false),
 		state(WebSocket::Idle),
 		errorCondition(ErrorGeneric),
@@ -221,6 +225,8 @@ public:
 		peerClosing(false),
 		peerCloseCode(-1),
 		disconnecting(false),
+		disconnectSent(false),
+		updateQueued(false),
 		retries(0)
 	{
 		if(!g_disconnectManager)
@@ -253,6 +259,7 @@ public:
 
 		updating = false;
 		disconnecting = false;
+		updateQueued = false;
 
 		delete req;
 		req = 0;
@@ -337,6 +344,18 @@ public:
 		update();
 	}
 
+	void refresh()
+	{
+		// only allow refresh requests if connected
+		if(state == Connected && !disconnecting)
+		{
+			if(!updating)
+				update();
+			else
+				updateQueued = true;
+		}
+	}
+
 private:
 	bool canReceive() const
 	{
@@ -365,7 +384,10 @@ private:
 	bool needUpdate() const
 	{
 		// always send this right away
-		if(disconnecting)
+		if(disconnecting && !disconnectSent)
+			return true;
+
+		if(updateQueued)
 			return true;
 
 		bool cscm = canSendCompleteMessage();
@@ -403,6 +425,8 @@ private:
 		if(updating)
 			return;
 
+		updateQueued = false;
+
 		updating = true;
 
 		keepAliveTimer->stop();
@@ -425,9 +449,10 @@ private:
 		{
 			events += WsEvent("OPEN");
 		}
-		else if(disconnecting)
+		else if(disconnecting && !disconnectSent)
 		{
 			events += WsEvent("DISCONNECT");
+			disconnectSent = true;
 		}
 		else
 		{
@@ -530,6 +555,7 @@ private:
 		if(connectPort != -1)
 			req->setConnectPort(connectPort);
 		req->setIgnorePolicies(ignorePolicies);
+		req->setTrustConnectHost(trustConnectHost);
 		req->setIgnoreTlsErrors(ignoreTlsErrors);
 		req->setSendBodyAfterAcknowledgement(true);
 
@@ -626,7 +652,8 @@ private slots:
 				QByteArray name = h.first.mid(9);
 				if(meta.contains(name))
 					meta.removeAll(name);
-				meta += HttpHeader(name, h.second);
+				if(!h.second.isEmpty())
+					meta += HttpHeader(name, h.second);
 			}
 		}
 
@@ -657,6 +684,10 @@ private slots:
 				return;
 			}
 
+			// correct the status code/reason
+			responseData.code = 101;
+			responseData.reason = "Switching Protocols";
+
 			// strip private headers from the initial response
 			responseData.headers.removeAll("Content-Length");
 			responseData.headers.removeAll("Content-Type");
@@ -672,7 +703,7 @@ private slots:
 			}
 		}
 
-		if(disconnecting)
+		if(disconnectSent)
 		{
 			cleanup();
 			emit q->disconnected();
@@ -911,6 +942,11 @@ void WebSocketOverHttp::setConnectionId(const QByteArray &id)
 	d->cid = id;
 }
 
+void WebSocketOverHttp::refresh()
+{
+	d->refresh();
+}
+
 void WebSocketOverHttp::clearDisconnectManager()
 {
 	delete g_disconnectManager;
@@ -943,6 +979,11 @@ void WebSocketOverHttp::setIgnorePolicies(bool on)
 	d->ignorePolicies = on;
 }
 
+void WebSocketOverHttp::setTrustConnectHost(bool on)
+{
+	d->trustConnectHost = on;
+}
+
 void WebSocketOverHttp::setIgnoreTlsErrors(bool on)
 {
 	d->ignoreTlsErrors = on;
