@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014 Fanout, Inc.
+ * Copyright (C) 2014-2017 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -42,6 +42,7 @@
 #include "testwebsocket.h"
 
 #define ACTIVITY_TIMEOUT 60000
+#define FRAME_SIZE_MAX 16384
 
 class HttpExtension
 {
@@ -223,6 +224,8 @@ public:
 		Closing
 	};
 
+	typedef QPair<WebSocket::Frame, bool> QueuedFrame;
+
 	WsProxySession *q;
 	State state;
 	ZRoutes *zroutes;
@@ -244,6 +247,7 @@ public:
 	WebSocket *inSock;
 	WebSocket *outSock;
 	int inPendingBytes;
+	QList<bool> inPendingFrames; // true means we should ack a send event
 	int outPendingBytes;
 	int outReadInProgress; // frame type or -1
 	QByteArray pathBeg;
@@ -258,6 +262,7 @@ public:
 	QDateTime activityTime;
 	QByteArray publicCid;
 	QTimer *keepAliveTimer;
+	QList<QueuedFrame> queuedInFrames; // frames to deliver after out read finishes
 
 	Private(WsProxySession *_q, ZRoutes *_zroutes, ConnectionManager *_connectionManager, StatsManager *_statsManager, WsControlManager *_wsControlManager) :
 		QObject(_q),
@@ -360,7 +365,7 @@ public:
 		}
 
 		if(!entry.asHost.isEmpty())
-			requestData.uri.setHost(entry.asHost);
+			ProxyUtil::applyHost(&requestData.uri, entry.asHost);
 
 		QByteArray path = requestData.uri.path(QUrl::FullyEncoded).toUtf8();
 
@@ -392,7 +397,13 @@ public:
 
 		log_debug("wsproxysession: %p %s has %d routes", q, qPrintable(host), targets.count());
 
-		bool trustedClient = ProxyUtil::manipulateRequestHeaders("wsproxysession", q, &requestData, defaultUpstreamKey, entry, sigIss, sigKey, acceptXForwardedProtocol, useXForwardedProtocol, xffTrustedRule, xffRule, origHeadersNeedMark, inSock->peerAddress(), InspectData());
+		foreach(const HttpHeader &h, entry.headers)
+		{
+			requestData.headers.removeAll(h.first);
+			requestData.headers += HttpHeader(h.first, h.second);
+		}
+
+		bool trustedClient = ProxyUtil::manipulateRequestHeaders("wsproxysession", q, &requestData, defaultUpstreamKey, entry, sigIss, sigKey, acceptXForwardedProtocol, useXForwardedProtocol, xffTrustedRule, xffRule, origHeadersNeedMark, inSock->peerAddress(), InspectData(), true);
 
 		// don't proxy extensions, as we may not know how to handle them
 		requestData.headers.removeAll("Sec-WebSocket-Extensions");
@@ -406,6 +417,14 @@ public:
 		tryNextTarget();
 	}
 
+	void writeInFrame(const WebSocket::Frame &frame, bool fromSendEvent = false)
+	{
+		inPendingBytes += frame.data.size();
+		inPendingFrames += fromSendEvent;
+
+		inSock->writeFrame(frame);
+	}
+
 	void tryNextTarget()
 	{
 		if(targets.isEmpty())
@@ -423,7 +442,7 @@ public:
 			uri.setScheme("ws");
 
 		if(!target.host.isEmpty())
-			uri.setHost(target.host);
+			ProxyUtil::applyHost(&uri, target.host);
 
 		subChannel = target.subChannel;
 
@@ -493,6 +512,9 @@ public:
 		if(target.trusted)
 			outSock->setIgnorePolicies(true);
 
+		if(target.trustConnectHost)
+			outSock->setTrustConnectHost(true);
+
 		if(target.insecure)
 			outSock->setIgnoreTlsErrors(true);
 
@@ -502,6 +524,8 @@ public:
 			outSock->setConnectPort(target.connectPort);
 		}
 
+		ProxyUtil::applyHostHeader(&requestData.headers, uri);
+
 		outSock->start(uri, requestData.headers);
 	}
 
@@ -544,7 +568,7 @@ public:
 
 			tryLogActivity();
 
-			if(detached)
+			if(detached && outReadInProgress == -1)
 				continue;
 
 			if(f.type == WebSocket::Frame::Text || f.type == WebSocket::Frame::Binary || f.type == WebSocket::Frame::Continuation)
@@ -566,28 +590,32 @@ public:
 						else
 							outReadInProgress = -1; // ignore rest of message
 					}
-					else if(f.type != WebSocket::Frame::Continuation && f.data.startsWith(messagePrefix))
+					else if(f.type != WebSocket::Frame::Continuation)
 					{
-						f.data = f.data.mid(messagePrefix.size());
-						inSock->writeFrame(f);
-						inPendingBytes += f.data.size();
+						if(f.data.startsWith(messagePrefix))
+						{
+							f.data = f.data.mid(messagePrefix.size());
+							writeInFrame(f);
 
-						restartKeepAlive();
+							restartKeepAlive();
+						}
+						else
+						{
+							log_debug("wsproxysession: dropping unprefixed message");
+						}
 					}
 					else if(f.type == WebSocket::Frame::Continuation)
 					{
 						assert(outReadInProgress != -1);
 
-						inSock->writeFrame(f);
-						inPendingBytes += f.data.size();
+						writeInFrame(f);
 
 						restartKeepAlive();
 					}
 				}
 				else
 				{
-					inSock->writeFrame(f);
-					inPendingBytes += f.data.size();
+					writeInFrame(f);
 
 					restartKeepAlive();
 				}
@@ -598,11 +626,16 @@ public:
 			else
 			{
 				// always relay non-content frames
-				inSock->writeFrame(f);
-				inPendingBytes += f.data.size();
+				writeInFrame(f);
 
 				restartKeepAlive();
 			}
+
+			if(outReadInProgress == -1 && !queuedInFrames.isEmpty())
+			{
+				foreach(const QueuedFrame &i, queuedInFrames)
+					writeInFrame(i.first, i.second);
+			}
 		}
 	}
 
@@ -683,6 +716,13 @@ private slots:
 
 		inPendingBytes -= contentBytes;
 
+		for(int n = 0; n < count; ++n)
+		{
+			bool fromSendEvent = inPendingFrames.takeFirst();
+			if(fromSendEvent)
+				wsControl->sendEventWritten();
+		}
+
 		if(!detached && outSock)
 			tryReadOut();
 	}
@@ -777,6 +817,7 @@ private slots:
 				connect(wsControl, &WsControlSession::closeEventReceived, this, &Private::wsControl_closeEventReceived);
 				connect(wsControl, &WsControlSession::detachEventReceived, this, &Private::wsControl_detachEventReceived);
 				connect(wsControl, &WsControlSession::cancelEventReceived, this, &Private::wsControl_cancelEventReceived);
+				connect(wsControl, &WsControlSession::error, this, &Private::wsControl_error);
 				wsControl->start(routeId, channelPrefix, inSock->requestUri());
 
 				if(!subChannel.isEmpty())
@@ -884,18 +925,65 @@ private slots:
 		}
 	}
 
-	void wsControl_sendEventReceived(WebSocket::Frame::Type type, const QByteArray &message)
+	void wsControl_sendEventReceived(WebSocket::Frame::Type type, const QByteArray &message, bool queue)
 	{
 		// this method accepts a full message, which must be typed
 		if(type == WebSocket::Frame::Continuation)
 			return;
 
-		// only send if we can, otherwise drop
-		if(inSock && inSock->canWrite())
+		// if we have no socket to write to, say the data was written anyway.
+		//   this is not quite correct but better than leaving the send event
+		//   dangling
+		if(!inSock)
 		{
-			inSock->writeFrame(WebSocket::Frame(type, message, false));
+			wsControl->sendEventWritten();
+			return;
+		}
 
-			inPendingBytes += message.size();
+		// if queue == false, drop if we can't send right now
+		if(!queue && (!inSock->canWrite() || outReadInProgress != -1))
+		{
+			// if drop is allowed, drop is success :)
+			wsControl->sendEventWritten();
+			return;
+		}
+
+		// split into frames to avoid credits issue
+		QList<WebSocket::Frame> frames;
+
+		if(message.size() > 0)
+		{
+			for(int n = 0; n < message.size(); n += FRAME_SIZE_MAX)
+			{
+				WebSocket::Frame::Type ftype;
+				if(n == 0)
+					ftype = type;
+				else
+					ftype = WebSocket::Frame::Continuation;
+
+				QByteArray data = message.mid(n, FRAME_SIZE_MAX);
+				bool more = (n + FRAME_SIZE_MAX < message.size());
+
+				frames += WebSocket::Frame(ftype, data, more);
+			}
+		}
+		else
+		{
+			frames += WebSocket::Frame(type, QByteArray(), false);
+		}
+
+		for(int n = 0; n < frames.count(); ++n)
+		{
+			bool fromSendEvent = (n + 1 >= frames.count());
+
+			if(outReadInProgress != -1)
+			{
+				queuedInFrames += QueuedFrame(frames[n], fromSendEvent);
+			}
+			else
+			{
+				writeInFrame(frames[n], fromSendEvent);
+			}
 		}
 
 		restartKeepAlive();
@@ -955,6 +1043,12 @@ private slots:
 		tryFinish();
 	}
 
+	void wsControl_error()
+	{
+		log_debug("wsproxysession: %p wscontrol session error", q);
+		wsControl_cancelEventReceived();
+	}
+
 	void keepAliveTimer_timeout()
 	{
 		wsControl->sendNeedKeepAlive();
@@ -982,6 +1076,16 @@ QByteArray WsProxySession::cid() const
 	return d->publicCid;
 }
 
+WebSocket *WsProxySession::inSocket() const
+{
+	return d->inSock;
+}
+
+WebSocket *WsProxySession::outSocket() const
+{
+	return d->outSock;
+}
+
 void WsProxySession::setDefaultSigKey(const QByteArray &iss, const QByteArray &key)
 {
 	d->defaultSigIss = iss;
